@@ -1,6 +1,6 @@
 /*
  * sar, sadc, sadf, mpstat and iostat common routines.
- * (C) 1999-2006 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1999-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -26,6 +26,7 @@
 #include <errno.h>
 #include <unistd.h>	/* For STDOUT_FILENO, among others */
 #include <dirent.h>
+#include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
@@ -55,7 +56,7 @@ unsigned int kb_shift;
 void print_version(void)
 {
    fprintf(stderr, _("sysstat version %s\n"), VERSION);
-   fprintf(stderr, "(C) Sebastien Godard (sysstat <at> wanadoo.fr)\n");
+   fprintf(stderr, "(C) Sebastien Godard (sysstat <at> orange.fr)\n");
    exit(1);
 }
 
@@ -128,8 +129,9 @@ int get_sys_cpu_nr(void)
    /* Get current file entry */
    while ((drd = readdir(dir)) != NULL) {
 
-      if (!strncmp(drd->d_name, "cpu", 3)) {
-	 sprintf(line, "%s/%s", SYSFS_DEVCPU, drd->d_name);
+      if (!strncmp(drd->d_name, "cpu", 3) && isdigit(drd->d_name[3])) {
+ 	 snprintf(line, MAX_PF_NAME, "%s/%s", SYSFS_DEVCPU, drd->d_name);
+	 line[MAX_PF_NAME - 1] = '\0';
 	 if (stat(line, &buf) < 0)
 	    continue;
 	 if (S_ISDIR(buf.st_mode))
@@ -179,7 +181,9 @@ int get_proc_cpu_nr(void)
  ***************************************************************************
  * Return the number of processors used on the machine:
  * 0: one proc and non SMP kernel
- * 1: one proc and SMP kernel, or SMP machine with all but one offlined CPU
+ * 1: one proc and SMP kernel (NB: on SMP machines where all the CPU's but
+ *    one have been disabled, we get the total number of proc since we use
+ *    /sys to count them).
  * 2: two proc...
  * Try to use /sys for that, or /proc/stat if /sys doesn't exist.
  ***************************************************************************
@@ -213,7 +217,8 @@ int get_dev_part_nr(char *dev_name)
    char dfile[MAX_PF_NAME], line[MAX_PF_NAME];
    int part = 0;
 
-   sprintf(dfile, "%s/%s", SYSFS_BLOCK, dev_name);
+   snprintf(dfile, MAX_PF_NAME, "%s/%s", SYSFS_BLOCK, dev_name);
+   dfile[MAX_PF_NAME - 1] = '\0';
 
    /* Open current device directory in /sys/block */
    if ((dir = opendir(dfile)) == NULL)
@@ -223,7 +228,8 @@ int get_dev_part_nr(char *dev_name)
    while ((drd = readdir(dir)) != NULL) {
       if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
 	 continue;
-      sprintf(line, "%s/%s/%s", dfile, drd->d_name, S_STAT);
+      snprintf(line, MAX_PF_NAME, "%s/%s/%s", dfile, drd->d_name, S_STAT);
+      line[MAX_PF_NAME - 1] = '\0';
 
       /* Try to guess if current entry is a directory containing a stat file */
       if (!access(line, R_OK))
@@ -261,7 +267,8 @@ int get_sysfs_dev_nr(int display_partitions)
    while ((drd = readdir(dir)) != NULL) {
       if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
 	 continue;
-      sprintf(line, "%s/%s/%s", SYSFS_BLOCK, drd->d_name, S_STAT);
+      snprintf(line, MAX_PF_NAME, "%s/%s/%s", SYSFS_BLOCK, drd->d_name, S_STAT);
+      line[MAX_PF_NAME - 1] = '\0';
 
       /* Try to guess if current entry is a directory containing a stat file */
       if (!access(line, R_OK)) {
@@ -432,19 +439,25 @@ int get_nfs_mount_nr(void)
 /*
  ***************************************************************************
  * Print banner
+ * Return code: 1 if S_TIME_FORMAT is set to ISO, or 0 otherwise.
  ***************************************************************************
  */
-inline void print_gal_header(struct tm *rectime, char *sysname, char *release, char *nodename)
+int print_gal_header(struct tm *rectime, char *sysname, char *release, char *nodename)
 {
    char cur_date[64];
    char *e;
+   int rc = 0;
 
-   if (((e = getenv(ENV_TIME_FMT)) != NULL) && !strcmp(e, K_ISO))
+   if (((e = getenv(ENV_TIME_FMT)) != NULL) && !strcmp(e, K_ISO)) {
       strftime(cur_date, sizeof(cur_date), "%Y-%m-%d", rectime);
+      rc = 1;
+   }
    else
       strftime(cur_date, sizeof(cur_date), "%x", rectime);
 
    printf("%s %s (%s) \t%s\n", sysname, release, nodename, cur_date);
+
+   return rc;
 }
 
 
@@ -469,7 +482,7 @@ void init_nls(void)
 
 /*
  ***************************************************************************
- * Get window height (number of lines)
+ * Get nb of rows and columns of current window
  ***************************************************************************
  */
 int get_win_height(void)
@@ -481,10 +494,10 @@ int get_win_height(void)
     */
    int rows = 3600 * 24;
 
-
-   if ((ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) != -1) && (win.ws_row > 2))
-      rows = win.ws_row - 2;
-
+   if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) != -1) {
+      if (win.ws_row > 2)
+	 rows = win.ws_row - 2;
+   }
    return rows;
 }
 
@@ -577,3 +590,105 @@ double ll_s_value(unsigned long long value1, unsigned long long value2,
    else
       return S_VALUE(value1, value2, itv);
 }
+
+
+/*
+ ***************************************************************************
+ * Compute time interval.
+ * The interval should always be smaller than 0xffffffff (ULONG_MAX on
+ * 32-bit architectures), except perhaps if it is the interval since
+ * system startup (we want stats since boot time).
+ * Interval is and'ed with mask 0xffffffff to handle overflow conditions
+ * that may happen since uptime values are unsigned long long but are
+ * calculated as a sum of values that _may_ be unsigned long only...
+ ***************************************************************************
+ */
+unsigned long long get_interval(unsigned long long prev_uptime,
+				unsigned long long curr_uptime)
+{
+   unsigned long long itv;
+
+   if (!prev_uptime)
+      itv = curr_uptime;
+   else
+      itv = (curr_uptime - prev_uptime) & 0xffffffff;
+   if (!itv)	/* Paranoia checking */
+      itv = 1;
+
+   return itv;
+}
+
+
+/*
+ ***************************************************************************
+ * Read machine uptime, independently of the number of processors
+ ***************************************************************************
+ */
+void readp_uptime(unsigned long long *uptime)
+{
+   FILE *fp;
+   char line[128];
+   unsigned long up_sec, up_cent;
+
+   if ((fp = fopen(UPTIME, "r")) == NULL)
+      return;
+
+   if (fgets(line, 128, fp) == NULL)
+      return;
+
+   sscanf(line, "%lu.%lu", &up_sec, &up_cent);
+
+   fclose(fp);
+
+   *uptime = up_sec * HZ + up_cent * HZ / 100;
+}
+
+
+/*
+ ***************************************************************************
+ * Read stats from /proc/meminfo
+ ***************************************************************************
+ */
+int readp_meminfo(struct meminf *st_mem)
+{
+   FILE *fp;
+   static char line[128];
+
+   if ((fp = fopen(MEMINFO, "r")) == NULL)
+      return 1;
+
+   while (fgets(line, 128, fp) != NULL) {
+
+      if (!strncmp(line, "MemTotal:", 9))
+	 /* Read the total amount of memory in kB */
+	 sscanf(line + 9, "%lu", &(st_mem->tlmkb));
+      else if (!strncmp(line, "MemFree:", 8))
+	 /* Read the amount of free memory in kB */
+	 sscanf(line + 8, "%lu", &(st_mem->frmkb));
+
+      else if (!strncmp(line, "Buffers:", 8))
+	 /* Read the amount of buffered memory in kB */
+	 sscanf(line + 8, "%lu", &(st_mem->bufkb));
+
+      else if (!strncmp(line, "Cached:", 7))
+	 /* Read the amount of cached memory in kB */
+	 sscanf(line + 7, "%lu", &(st_mem->camkb));
+
+      else if (!strncmp(line, "SwapCached:", 11))
+	 /* Read the amount of cached swap in kB */
+	 sscanf(line + 11, "%lu", &(st_mem->caskb));
+
+      else if (!strncmp(line, "SwapTotal:", 10))
+	 /* Read the total amount of swap memory in kB */
+	 sscanf(line + 10, "%lu", &(st_mem->tlskb));
+
+      else if (!strncmp(line, "SwapFree:", 9))
+	 /* Read the amount of free swap memory in kB */
+	 sscanf(line + 9, "%lu", &(st_mem->frskb));
+   }
+
+   fclose(fp);
+
+   return 0;
+}
+
