@@ -1,6 +1,6 @@
 /*
  * iostat: report CPU and I/O statistics
- * (C) 1998-2006 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1998-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -33,11 +33,11 @@
 #include <sys/stat.h>
 #include <sys/utsname.h>
 
+#include "version.h"
 #include "iostat.h"
 #include "common.h"
 #include "ioconf.h"
 
-
 #ifdef USE_NLS
 #include <locale.h>
 #include <libintl.h>
@@ -46,6 +46,8 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)sysstat-" VERSION ": " __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
 
 struct comm_stats  comm_stats[2];
 struct io_stats *st_iodev[2];
@@ -53,19 +55,13 @@ struct io_nfs_stats *st_ionfs[2];
 struct io_hdr_stats *st_hdr_iodev, *st_hdr_ionfs;
 struct io_dlist *st_dev_list;
 
-/* Nb of devices and partitions found */
-int iodev_nr = 0;
-/* Nb of NFS mounted directories found */
-int ionfs_nr = 0;
-
-/* Nb of devices entered on the command line */
-int dlist_idx = 0;
+int iodev_nr = 0;	/* Nb of devices and partitions found */
+int ionfs_nr = 0;	/* Nb of NFS mounted directories found */
+int cpu_nr = 0;		/* Nb of processors on the machine */
+int dlist_idx = 0;	/* Nb of devices entered on the command line */
 
 long interval = 0;
-unsigned char timestamp[64];
-
-/* Nb of processors on the machine */
-int cpu_nr = 0;
+char timestamp[64];
 
 
 /*
@@ -77,7 +73,7 @@ void usage(char *progname)
 {
    fprintf(stderr, _("Usage: %s [ options... ] [ <interval> [ <count> ] ]\n"
 		   "Options are:\n"
-		   "[ -c ] [ -d ] [ -n ] [ -k | -m ] [ -t ] [ -V ] [ -x ]\n"
+		   "[ -c ] [ -d ] [ -N ] [ -n ] [ -k | -m ] [ -t ] [ -V ] [ -x ]\n"
 		   "[ <device> [ ... ] | ALL ] [ -p [ <device> | ALL ] ]\n"),
 	   progname);
    exit(1);
@@ -125,7 +121,7 @@ void set_entries_inactive(int ioln_nr, struct io_hdr_stats *st_hdr_ioln)
 
 /*
  ***************************************************************************
- * Set structures's state to free for inactive entries
+ * Set structures' state to free for inactive entries
  ***************************************************************************
  */
 void free_inactive_entries(int ioln_nr, struct io_hdr_stats *st_hdr_ioln)
@@ -222,7 +218,7 @@ int update_dev_list(int *dlist_idx, char *device_name)
    }
 
    if (i == *dlist_idx) {
-      /* Device not found: store it */
+      /* Device not found: Store it */
       (*dlist_idx)++;
       strncpy(sdli->dev_name, device_name, MAX_NAME_LEN - 1);
    }
@@ -395,7 +391,6 @@ void read_proc_stat(int curr, int flags)
    unsigned long long cc_idle, cc_iowait, cc_steal;
    unsigned long long cc_user, cc_nice, cc_system, cc_hardirq, cc_softirq;
 
-
    /*
     * Prepare pointers on the 4 disk structures in case we have a
     * /proc/stat file with "disk_rblk", etc. entries.
@@ -445,12 +440,11 @@ void read_proc_stat(int curr, int flags)
 	    			   comm_stats[curr].cpu_steal;
       }
 
-      else if ((!strncmp(line, "cpu0", 4)) && (cpu_nr > 1)) {
+      else if ((!strncmp(line, "cpu0", 4)) && !comm_stats[curr].uptime0
+	       && (cpu_nr > 1)) {
 	 /*
 	  * Read CPU line for proc#0 (if available).
-	  * This is necessary to compute time interval since
-	  * processors may be disabled (offline) sometimes.
-	  * (Assume that proc#0 can never be offline).
+	  * Done only if /proc/uptime was unavailable.
 	  */
 	 cc_iowait = cc_hardirq = cc_softirq = cc_steal = 0;
 	 sscanf(line + 5, "%llu %llu %llu %llu %llu %llu %llu %llu",
@@ -588,7 +582,8 @@ void read_sysfs_dlist_part_stat(int curr, char *dev_name)
    struct dirent *drd;
    char dfile[MAX_PF_NAME], filename[MAX_PF_NAME];
 
-   sprintf(dfile, "%s/%s", SYSFS_BLOCK, dev_name);
+   snprintf(dfile, MAX_PF_NAME, "%s/%s", SYSFS_BLOCK, dev_name);
+   dfile[MAX_PF_NAME - 1] = '\0';
 
    /* Open current device directory in /sys/block */
    if ((dir = opendir(dfile)) == NULL)
@@ -598,7 +593,8 @@ void read_sysfs_dlist_part_stat(int curr, char *dev_name)
    while ((drd = readdir(dir)) != NULL) {
       if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
 	 continue;
-      sprintf(filename, "%s/%s/%s", dfile, drd->d_name, S_STAT);
+      snprintf(filename, MAX_PF_NAME, "%s/%s/%s", dfile, drd->d_name, S_STAT);
+      filename[MAX_PF_NAME - 1] = '\0';
 
       /* Read current partition stats */
       read_sysfs_file_stat(curr, filename, drd->d_name, DT_PARTITION);
@@ -626,8 +622,9 @@ void read_sysfs_dlist_stat(int curr, int flags)
 
    for (dev = 0; dev < dlist_idx; dev++) {
       st_dev_list_i = st_dev_list + dev;
-      sprintf(filename, "%s/%s/%s",
+      snprintf(filename, MAX_PF_NAME, "%s/%s/%s",
 	      SYSFS_BLOCK, st_dev_list_i->dev_name, S_STAT);
+      filename[MAX_PF_NAME - 1] = '\0';
 
       /* Read device stats */
       ok = read_sysfs_file_stat(curr, filename, st_dev_list_i->dev_name, DT_DEVICE);
@@ -665,7 +662,8 @@ void read_sysfs_stat(int curr, int flags)
       while ((drd = readdir(dir)) != NULL) {
 	 if (!strcmp(drd->d_name, ".") || !strcmp(drd->d_name, ".."))
 	    continue;
-	 sprintf(filename, "%s/%s/%s", SYSFS_BLOCK, drd->d_name, S_STAT);
+	 snprintf(filename, MAX_PF_NAME, "%s/%s/%s", SYSFS_BLOCK, drd->d_name, S_STAT);
+	 filename[MAX_PF_NAME - 1] = '\0';
 	
 	 /* If current entry is a directory, try to read its stat file */
 	 ok = read_sysfs_file_stat(curr, filename, drd->d_name, DT_DEVICE);
@@ -696,6 +694,7 @@ void read_diskstats_stat(int curr, int flags)
 {
    FILE *fp;
    char line[256], dev_name[MAX_NAME_LEN];
+   char *dm_name;
    struct io_stats sdev;
    int i;
    unsigned long rd_ios, rd_merges_or_rd_sec, rd_ticks_or_wr_sec, wr_ios;
@@ -743,7 +742,7 @@ void read_diskstats_stat(int curr, int flags)
 	 sdev.wr_sectors = rd_ticks_or_wr_sec;
       }
       else
-	 /* Unknown entry: ignore it */
+	 /* Unknown entry: Ignore it */
 	 continue;
 
       if ((ioc_dname = ioc_name(major, minor)) != NULL) {
@@ -755,9 +754,18 @@ void read_diskstats_stat(int curr, int flags)
 	    strncpy(dev_name, ioc_dname, MAX_NAME_LEN);
       }
 
+      if ((DISPLAY_DEVMAP_NAME(flags)) && (major == DEVMAP_MAJOR)) {
+	 /*
+	  * If the device is a device mapper device, try to get its
+	  * assigned name of its logical device.
+	  */
+	 dm_name = transform_devmapname(major, minor);
+	 if (dm_name)
+	    strcpy(dev_name, dm_name);
+      }
+
       save_stats(dev_name, curr, &sdev, iodev_nr, st_hdr_iodev);
    }
-
    fclose(fp);
 
    /* Free structures corresponding to unregistered devices */
@@ -778,7 +786,7 @@ void read_ppartitions_stat(int curr, int flags)
    unsigned long rd_ios, rd_merges, rd_ticks, wr_ios, wr_merges, wr_ticks;
    unsigned long ios_pgr, tot_ticks, rq_ticks;
    unsigned long long rd_sec, wr_sec;
-   char *ioc_dname;
+   char *ioc_dname, *dm_name;
    unsigned int major, minor;
 
    /* Every I/O device entry is potentially unregistered */
@@ -803,7 +811,7 @@ void read_ppartitions_stat(int curr, int flags)
 	 sdev.rq_ticks   = rq_ticks;
       }
       else
-	 /* Unknown entry: ignore it */
+	 /* Unknown entry: Ignore it */
 	 continue;
 
       if ((ioc_dname = ioc_name(major, minor)) != NULL) {
@@ -812,9 +820,15 @@ void read_ppartitions_stat(int curr, int flags)
 	    strncpy(dev_name, ioc_dname, MAX_NAME_LEN);
       }
 
+      if ((DISPLAY_DEVMAP_NAME(flags)) && (major == DEVMAP_MAJOR)) {
+	 /* Get device mapper logical name */
+	 dm_name = transform_devmapname(major, minor);
+	 if (dm_name)
+	    strcpy(dev_name, dm_name);
+      }
+
       save_stats(dev_name, curr, &sdev, iodev_nr, st_hdr_iodev);
    }
-
    fclose(fp);
 
    /* Free structures corresponding to unregistered devices */
@@ -983,7 +997,7 @@ void write_ext_stat(int curr, unsigned long long itv, int flags, int fctr,
    util = S_VALUE(ioj->tot_ticks, ioi->tot_ticks, itv);
    svctm = tput ? util / tput : 0.0;
    /*
-    * kernel gives ticks already in milliseconds for all platforms
+    * Kernel gives ticks already in milliseconds for all platforms
     * => no need for further scaling.
     */
    await = nr_ios ?
@@ -1013,7 +1027,7 @@ void write_ext_stat(int curr, unsigned long long itv, int flags, int fctr,
 	  await,
 	  /* The ticks output is biased to output 1000 ticks per second */
 	  svctm,
-	  /* Again: ticks in milliseconds */
+	  /* Again: Ticks in milliseconds */
 	  util / 10.0);
 }
 
@@ -1086,63 +1100,35 @@ void write_nfs_stat(int curr, unsigned long long itv, int flags, int fctr,
  * Print everything now (stats and uptime)
  ***************************************************************************
  */
-int write_stat(int curr, int flags, struct tm *rectime)
+int write_stats(int curr, int flags, struct tm *rectime)
 {
    int dev, i, fctr = 1;
    unsigned long long itv;
    struct io_hdr_stats *shi;
    struct io_dlist *st_dev_list_i;
 
-   /*
-    * Under very special circumstances, STDOUT may become unavailable,
-    * This is what we try to guess here
-    */
-   if (write(STDOUT_FILENO, "", 0) == -1) {
-      perror("stdout");
-      exit(6);
-   }
+   /* Test stdout */
+   TEST_STDOUT(STDOUT_FILENO);
 
    /* Print time stamp */
    if (DISPLAY_TIMESTAMP(flags)) {
-      strftime(timestamp, sizeof(timestamp), "%X", rectime);
+      if (DISPLAY_ISO(flags))
+	 strftime(timestamp, sizeof(timestamp), "%FT%T%z", rectime);
+      else
+	 strftime(timestamp, sizeof(timestamp), "%X", rectime);
       printf(_("Time: %s\n"), timestamp);
    }
 
-   /*
-    * itv is multiplied by the number of processors.
-    * This is OK to compute CPU usage since the number of jiffies spent in the
-    * different modes (user, nice, etc.) is the sum of all the processors.
-    * But itv should be reduced to one processor before displaying disk or NFS
-    * utilization.
-    */
-   if (!comm_stats[!curr].uptime)
-      /*
-       * This is the first report displaying stats since system startup.
-       * Only in this case we admit that the interval may be greater
-       * than 0xffffffff, else it was an overflow.
-       */
-      itv = comm_stats[curr].uptime;
-   else
-      /* uptime in jiffies */
-      itv = (comm_stats[curr].uptime - comm_stats[!curr].uptime)
-	 & 0xffffffff;
-   if (!itv)
-      itv = 1;
+   /* Interval is multiplied by the number of processors */
+   itv = get_interval(comm_stats[!curr].uptime, comm_stats[curr].uptime);
 
    if (DISPLAY_CPU(flags))
       /* Display CPU utilization */
       write_cpu_stat(curr, itv);
 
-   if (cpu_nr > 1) {
+   if (cpu_nr > 1)
       /* On SMP machines, reduce itv to one processor (see note above) */
-      if (!comm_stats[!curr].uptime0)
-	 itv = comm_stats[curr].uptime0;
-      else
-	 itv = (comm_stats[curr].uptime0 - comm_stats[!curr].uptime0)
-	    & 0xffffffff;
-      if (!itv)
-	 itv = 1;
-   }
+      itv = get_interval(comm_stats[!curr].uptime0, comm_stats[curr].uptime0);
 
    if (DISPLAY_DISK(flags)) {
       struct io_stats *ioi, *ioj;
@@ -1165,7 +1151,7 @@ int write_stat(int curr, int flags, struct tm *rectime)
 	    if (dlist_idx && !HAS_SYSFS(flags)) {
 	       /*
 		* With sysfs, only stats for the requested devices are read.
-		* With /proc/{diskstats,partitions}, stats for every devices
+		* With /proc/{diskstats,partitions}, stats for every device
 		* are read. Thus we need to check if stats for current device
 		* are to be displayed.
 		*/
@@ -1175,7 +1161,7 @@ int write_stat(int curr, int flags, struct tm *rectime)
 		     break;
 	       }
 	       if (dev == dlist_idx)
-		  /* Device not found in list: don't display it */
+		  /* Device not found in list: Don't display it */
 		  continue;
 	    }
 	
@@ -1233,8 +1219,7 @@ int write_stat(int curr, int flags, struct tm *rectime)
 
 /*
  ***************************************************************************
- * Main loop: read I/O stats from the relevant sources,
- * and display them.
+ * Main loop: Read I/O stats from the relevant sources and display them
  ***************************************************************************
  */
 void rw_io_stat_loop(int flags, long int count, struct tm *rectime)
@@ -1243,6 +1228,16 @@ void rw_io_stat_loop(int flags, long int count, struct tm *rectime)
    int next;
 
    do {
+      if (cpu_nr > 1) {
+	 /*
+	  * Read system uptime (only for SMP machines).
+	  * Init uptime0. So if /proc/uptime cannot fill it, this will be
+	  * done by /proc/stat.
+	  */
+	 comm_stats[curr].uptime0 = 0;
+	 readp_uptime(&(comm_stats[curr].uptime0));
+      }
+
       /* Read kernel statistics (CPU, and possibly disks for old kernels) */
       read_proc_stat(curr, flags);
 
@@ -1275,11 +1270,11 @@ void rw_io_stat_loop(int flags, long int count, struct tm *rectime)
       if (HAS_NFS(flags))
 	 read_nfs_stat(curr, flags);
 
-      /* Save time */
+      /* Get time */
       get_localtime(rectime);
 
       /* Print results */
-      if ((next = write_stat(curr, flags, rectime))
+      if ((next = write_stats(curr, flags, rectime))
 	  && (count > 0))
 	 count--;
       fflush(stdout);
@@ -1370,6 +1365,10 @@ int main(int argc, char **argv)
 	       flags |= I_D_MEGABYTES;	/* Display stats in MB/s */
 	       break;
 	
+	     case 'N':
+	       flags |= I_D_DEVMAP_NAME; /* Display device mapper logical name */
+	       break;
+	
 	     case 'n':
 	       flags |= I_D_NFS;	/* Display NFS stats */
 	       break;
@@ -1410,11 +1409,14 @@ int main(int argc, char **argv)
 	 it = 1;
       }
 
-      else {
+      else if (it > 0) {
 	 count = atol(argv[opt++]);
 	 if (count < 1)
 	   usage(argv[0]);
+	 it = -1;
       }
+      else
+	 usage(argv[0]);
    }
 
    /* Default: Display CPU and DISK reports */
@@ -1445,8 +1447,9 @@ int main(int argc, char **argv)
 
    /* Get system name, release number and hostname */
    uname(&header);
-   print_gal_header(&rectime,
-		    header.sysname, header.release, header.nodename);
+   if (print_gal_header(&rectime,
+			header.sysname, header.release, header.nodename))
+      flags |= I_D_ISO;
    printf("\n");
 
    /* Set a handler for SIGALRM */
