@@ -1,6 +1,6 @@
 /*
  * sadf: system activity data formatter
- * (C) 1999-2006 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1999-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -27,9 +27,11 @@
 #include <time.h>
 #include <errno.h>
 
+#include "version.h"
 #include "sadf.h"
 #include "sa.h"
 #include "common.h"
+#include "ioconf.h"
 
 
 #ifdef USE_NLS
@@ -40,24 +42,26 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)sysstat-" VERSION ": " __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
 
 long interval = -1, count = 0;
 unsigned int sadf_actflag = 0;
 unsigned int flags = 0;
-unsigned short format = 0;	/* Output format */
+unsigned int format = 0;	/* Output format */
 unsigned char irq_bitmap[(NR_IRQS / 8) + 1];
 unsigned char cpu_bitmap[(NR_CPUS / 8) + 1];
 
 struct file_hdr file_hdr;
-struct file_stats file_stats[DIM];
-struct stats_one_cpu *st_cpu[DIM];
-struct stats_serial *st_serial[DIM];
-struct stats_irq_cpu *st_irq_cpu[DIM];
-struct stats_net_dev *st_net_dev[DIM];
-struct disk_stats *st_disk[DIM];
+struct file_stats file_stats[3];
+struct stats_one_cpu *st_cpu[3];
+struct stats_serial *st_serial[3];
+struct stats_irq_cpu *st_irq_cpu[3];
+struct stats_net_dev *st_net_dev[3];
+struct disk_stats *st_disk[3];
 
 /* Array members of common types are always packed */
-unsigned int interrupts[DIM][NR_IRQS];
+unsigned int interrupts[3][NR_IRQS];
 
 struct tm rectime, loctime;
 /* Contain the date specified by -s and -e options */
@@ -100,7 +104,7 @@ void prtab(int nr_tab)
  * printf() function modified for XML display
  ***************************************************************************
  */
-void xprintf(short nr_tab, const char *fmt, ...)
+void xprintf(int nr_tab, const char *fmt, ...)
 {
    static char buf[1024];
    va_list args;
@@ -122,17 +126,17 @@ void xprintf(short nr_tab, const char *fmt, ...)
  * which is the number of seconds since 1970 _always_ expressed in UTC.
  ***************************************************************************
 */
-void set_rectime(short curr)
+void set_rectime(int curr)
 {
    struct tm *ltm;
 
-   ltm = localtime(&file_stats[curr].ust_time);
+   ltm = localtime((const time_t *) &file_stats[curr].ust_time);
    loctime = *ltm;
 
    if (!PRINT_TRUE_TIME(flags) ||
        ((format != S_O_DB_OPTION) && (format != S_O_XML_OPTION)))
       /* Option -t is valid only with options -d and -x */
-      ltm = gmtime(&file_stats[curr].ust_time);
+      ltm = gmtime((const time_t *) &file_stats[curr].ust_time);
 
    rectime = *ltm;
 }
@@ -143,7 +147,7 @@ void set_rectime(short curr)
  * Set timestamp string
  ***************************************************************************
 */
-void set_timestamp(short curr, char *cur_time, int len)
+void set_timestamp(int curr, char *cur_time, int len)
 {
    set_rectime(curr);
 	
@@ -271,7 +275,7 @@ static void render(int isdb, char *pre, int rflags, const char *pptxt,
  * making it easier for them to remain in sync and print the same data.
  ***************************************************************************
  */
-void write_mech_stats(short curr, unsigned int act,
+void write_mech_stats(int curr, unsigned int act,
 		      unsigned long dt, unsigned long long itv,
 		      unsigned long long g_itv, char *cur_time)
 {
@@ -288,6 +292,7 @@ void write_mech_stats(short curr, unsigned int act,
     */
    snprintf(pre, 80, "%s%s%ld%s%s",
 	    file_hdr.sa_nodename, seps[isdb], dt, seps[isdb], cur_time);
+   pre[79] = '\0';
 
    if (GET_PROC(act)) {
       /* The first one as an example */
@@ -450,10 +455,38 @@ void write_mech_stats(short curr, unsigned int act,
 	     NOVAL,
 	     S_VALUE(fsj->pgfault, fsi->pgfault, itv));
 
-      render(isdb, pre, PT_NEWLIN,
+      render(isdb, pre, PT_NOFLAG,
 	     "-\tmajflt/s", NULL, NULL,
 	     NOVAL,
 	     S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv));
+
+      render(isdb, pre, PT_NOFLAG,
+	     "-\tpgfree/s", NULL, NULL,
+	     NOVAL,
+	     S_VALUE(fsj->pgfree, fsi->pgfree, itv));
+
+      render(isdb, pre, PT_NOFLAG,
+	     "-\tpgscank/s", NULL, NULL,
+	     NOVAL,
+	     S_VALUE(fsj->pgscan_kswapd, fsi->pgscan_kswapd, itv));
+
+      render(isdb, pre, PT_NOFLAG,
+	     "-\tpgscand/s", NULL, NULL,
+	     NOVAL,
+	     S_VALUE(fsj->pgscan_direct, fsi->pgscan_direct, itv));
+
+      render(isdb, pre, PT_NOFLAG,
+	     "-\tpgsteal/s", NULL, NULL,
+	     NOVAL,
+	     S_VALUE(fsj->pgsteal, fsi->pgsteal, itv));
+
+      render(isdb, pre, PT_NEWLIN,
+	     "-\t%%vmeff", NULL, NULL,
+	     NOVAL,
+	     (fsi->pgscan_kswapd + fsi->pgscan_direct - fsj->pgscan_kswapd - fsj->pgscan_direct) ?
+	     SP_VALUE(fsj->pgsteal, fsi->pgsteal,
+		      fsi->pgscan_kswapd + fsi->pgscan_direct -
+		      fsj->pgscan_kswapd - fsj->pgscan_direct) : 0.0);
    }
 
    /* Print number of swap pages brought in and out */
@@ -494,18 +527,18 @@ void write_mech_stats(short curr, unsigned int act,
       render(isdb, pre, PT_NOFLAG,
 	     "-\tfrmpg/s", NULL, NULL,
 	     NOVAL,
-	     S_VALUE((double) PG(fsj->frmkb),
-		     (double) PG(fsi->frmkb), itv));
+	     S_VALUE((double) KB_TO_PG(fsj->frmkb),
+		     (double) KB_TO_PG(fsi->frmkb), itv));
 
       render(isdb, pre, PT_NOFLAG,
 	     "-\tbufpg/s", NULL, NULL,
-	     NOVAL, S_VALUE((double) PG(fsj->bufkb),
-			    (double) PG(fsi->bufkb), itv));
+	     NOVAL, S_VALUE((double) KB_TO_PG(fsj->bufkb),
+			    (double) KB_TO_PG(fsi->bufkb), itv));
 
       render(isdb, pre, PT_NEWLIN,
 	     "-\tcampg/s", NULL, NULL,
-	     NOVAL, S_VALUE((double) PG(fsj->camkb),
-			    (double) PG(fsi->camkb), itv));
+	     NOVAL, S_VALUE((double) KB_TO_PG(fsj->camkb),
+			    (double) KB_TO_PG(fsi->camkb), itv));
    }
 
    /* Print TTY statistics (serial lines) */
@@ -638,45 +671,16 @@ void write_mech_stats(short curr, unsigned int act,
 	     fsi->dentry_stat, DNOVAL);
 
       render(isdb, pre, PT_USEINT,
-	     "-\tfile-sz", NULL, NULL,
+	     "-\tfile-nr", NULL, NULL,
 	     fsi->file_used, DNOVAL);
 
       render(isdb, pre, PT_USEINT,
-	     "-\tinode-sz", NULL, NULL,
+	     "-\tinode-nr", NULL, NULL,
 	     fsi->inode_used, DNOVAL);
 
-      render(isdb, pre, PT_USEINT,
-	     "-\tsuper-sz", NULL, NULL,
-	     fsi->super_used, DNOVAL);
-
-      render(isdb, pre, PT_NOFLAG,
-	     "-\t%%super-sz", NULL, NULL,
-	     NOVAL,
-	     fsi->super_max
-	     ? ((double) (fsi->super_used * 100)) / fsi->super_max
-	     : 0.0);
-			
-      render(isdb, pre, PT_USEINT,
-	     "-\tdquot-sz", NULL, NULL,
-	     fsi->dquot_used, DNOVAL);
-
-      render(isdb, pre, PT_NOFLAG,
-	     "-\t%%dquot-sz", NULL, NULL,
-	     NOVAL,
-	     fsi->dquot_max
-	     ? ((double) (fsi->dquot_used * 100)) / fsi->dquot_max
-	     : 0.0);
-
-      render(isdb, pre, PT_USEINT,
-	     "-\trtsig-sz", NULL, NULL,
-	     fsi->rtsig_queued, DNOVAL);
-
-      render(isdb, pre, PT_NEWLIN,
-	     "-\t%%rtsig-sz", NULL, NULL,
-	     NOVAL,
-	     fsi->rtsig_max
-	     ? ((double) (fsi->rtsig_queued * 100)) / fsi->rtsig_max
-	     : 0.0);
+      render(isdb, pre, PT_USEINT | PT_NEWLIN,
+	     "-\tpty-nr", NULL, NULL,
+	     fsi->pty_nr, DNOVAL);
    }
 
    /* Print network interface statistics */
@@ -705,12 +709,12 @@ void write_mech_stats(short curr, unsigned int act,
 		NOVAL, S_VALUE(sndj->tx_packets, sndi->tx_packets, itv));
 
 	 render(isdb, pre, PT_NOFLAG,
-		"%s\trxbyt/s", NULL, cons(sv, ifc, NULL),
-		NOVAL, S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv));
+		"%s\trxkB/s", NULL, cons(sv, ifc, NULL),
+		NOVAL, S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv) / 1024);
 
 	 render(isdb, pre, PT_NOFLAG,
-		"%s\ttxbyt/s", NULL, cons(sv, ifc, NULL),
-		NOVAL, S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv));
+		"%s\ttxkB/s", NULL, cons(sv, ifc, NULL),
+		NOVAL, S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv) / 1024);
 
 	 render(isdb, pre, PT_NOFLAG,
 		"%s\trxcmp/s", NULL, cons(sv, ifc, NULL),
@@ -798,8 +802,11 @@ void write_mech_stats(short curr, unsigned int act,
       render(isdb, pre, PT_USEINT,
 	     "-\trawsck", NULL, NULL, fsi->raw_inuse, DNOVAL);
 
-      render(isdb, pre, PT_USEINT | PT_NEWLIN,
+      render(isdb, pre, PT_USEINT,
 	     "-\tip-frag", NULL, NULL, fsi->frag_inuse, DNOVAL);
+
+      render(isdb, pre, PT_USEINT | PT_NEWLIN,
+	     "-\ttcp-tw", NULL, NULL, fsi->tcp_tw, DNOVAL);
    }
 
    /* Print load averages and queue length */
@@ -839,7 +846,13 @@ void write_mech_stats(short curr, unsigned int act,
 
 	 j = check_disk_reg(&file_hdr, st_disk, curr, !curr, i);
 	 sdj = st_disk[!curr] + j;
-	 name = get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags));
+
+	 name = NULL;
+	 if ((USE_PRETTY_OPTION(flags)) && (sdi->major == DEVMAP_MAJOR))
+	    name = transform_devmapname(sdi->major, sdi->minor);
+	
+	 if (!name)
+	    name = get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags));
 
 	 tput = ((double) (sdi->nr_ios - sdj->nr_ios)) * HZ / itv;
 	 util = S_VALUE(sdj->tot_ticks, sdi->tot_ticks, itv);
@@ -959,14 +972,15 @@ void write_mech_stats(short curr, unsigned int act,
  * Write system statistics
  ***************************************************************************
  */
-int write_parsable_stats(short curr, unsigned int act, int reset, long *cnt,
+int write_parsable_stats(int curr, unsigned int act, int reset, long *cnt,
 			 int use_tm_start, int use_tm_end)
 {
    unsigned long long dt, itv, g_itv;
    char cur_time[26];
+   static int cross_day = 0;
 
    /* Check time (1) */
-   if (!next_slice(file_stats[2].uptime, file_stats[curr].uptime, &file_hdr,
+   if (!next_slice(file_stats[2].uptime0, file_stats[curr].uptime0,
 		   reset, interval))
       /* Not close enough to desired interval */
       return 0;
@@ -974,6 +988,20 @@ int write_parsable_stats(short curr, unsigned int act, int reset, long *cnt,
    /* Set current timestamp */
    set_timestamp(curr, cur_time, 26);
 
+   /* Check if we are beginning a new day */
+   if (use_tm_start && file_stats[!curr].ust_time &&
+       (file_stats[curr].ust_time > file_stats[!curr].ust_time) &&
+       (file_stats[curr].hour < file_stats[!curr].hour))
+      cross_day = 1;
+
+   if (cross_day)
+      /*
+       * This is necessary if we want to properly handle something like:
+       * sar -s time_start -e time_end with
+       * time_start(day D) > time_end(day D+1)
+       */
+      loctime.tm_hour +=24;
+
    /* Check time (2) */
    if (use_tm_start && (datecmp(&loctime, &tm_start) < 0))
      /* it's too soon... */
@@ -1006,7 +1034,7 @@ int write_parsable_stats(short curr, unsigned int act, int reset, long *cnt,
  * Display XML activity records
  ***************************************************************************
  */
-void write_xml_stats(short curr, short *tab)
+void write_xml_stats(int curr, int *tab)
 {
    int i, j, k;
    unsigned long long dt, itv, g_itv;
@@ -1173,6 +1201,7 @@ void write_xml_stats(short curr, short *tab)
       struct disk_stats
 	 *sdi = st_disk[curr],
 	 *sdj;
+      char *name = NULL;
 
       xprintf(*tab, "<disk per=\"second\">");
       (*tab)++;
@@ -1195,11 +1224,16 @@ void write_xml_stats(short curr, short *tab)
 	    ((sdi->rd_sect - sdj->rd_sect) + (sdi->wr_sect - sdj->wr_sect)) /
 	    ((double) (sdi->nr_ios - sdj->nr_ios)) : 0.0;
 
+	 if ((USE_PRETTY_OPTION(flags)) && (sdi->major == DEVMAP_MAJOR))
+	    name = transform_devmapname(sdi->major, sdi->minor);
+	
+	 if (!name)
+	    name = get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags));
+
 	 xprintf(*tab, "<disk-device dev=\"%s\" tps=\"%.2f\" rd_sec=\"%.2f\" "
 		       "wr_sec=\"%.2f\" avgrq-sz=\"%.2f\" avgqu-sz=\"%.2f\" "
 		       "await=\"%.2f\" svctm=\"%.2f\" util-percent=\"%.2f\"/>",
-		 /* Confusion possible here between index and minor numbers */
-		 get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags)),
+		 name,
 		 S_VALUE(sdj->nr_ios,  sdi->nr_ios,  itv),
 		 ll_s_value(sdj->rd_sect, sdi->rd_sect, itv),
 		 ll_s_value(sdj->wr_sect, sdi->wr_sect, itv),
@@ -1260,12 +1294,12 @@ void write_xml_stats(short curr, short *tab)
 	
 	 xprintf((*tab)++, "<net-device iface=\"%s\">",
 		 sndi->interface);
-	 xprintf(*tab, "<net-dev rxpck=\"%.2f\" txpck=\"%.2f\" rxbyt=\"%.2f\" "
-		       "txbyt=\"%.2f\" rxcmp=\"%.2f\" txcmp=\"%.2f\" rxmcst=\"%.2f\"/>",
+	 xprintf(*tab, "<net-dev rxpck=\"%.2f\" txpck=\"%.2f\" rxkB=\"%.2f\" "
+		       "txkB=\"%.2f\" rxcmp=\"%.2f\" txcmp=\"%.2f\" rxmcst=\"%.2f\"/>",
 		 S_VALUE(sndj->rx_packets, sndi->rx_packets, itv),
 		 S_VALUE(sndj->tx_packets, sndi->tx_packets, itv),
-		 S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv),
-		 S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv),
+		 S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv) / 1024,
+		 S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv) / 1024,
 		 S_VALUE(sndj->rx_compressed, sndi->rx_compressed, itv),
 		 S_VALUE(sndj->tx_compressed, sndi->tx_compressed, itv),
 		 S_VALUE(sndj->multicast, sndi->multicast, itv));
@@ -1309,19 +1343,29 @@ void write_xml_stats(short curr, short *tab)
 	   S_VALUE(fsj->nfsd_getattcnt, fsi->nfsd_getattcnt, itv));
 
    xprintf(*tab, "<net-sock totsck=\"%u\" tcpsck=\"%u\" udpsck=\"%u\" "
-	         "rawsck=\"%u\" ip-frag=\"%u\"/>",
+	         "rawsck=\"%u\" ip-frag=\"%u\" tcp-tw=\"%u\"/>",
 	   fsi->sock_inuse, fsi->tcp_inuse, fsi->udp_inuse,
-	   fsi->raw_inuse, fsi->frag_inuse);
+	   fsi->raw_inuse, fsi->frag_inuse, fsi->tcp_tw);
 
    xprintf(--(*tab), "</network>");
 
    /* paging */
    xprintf(*tab, "<paging per=\"second\" pgpgin=\"%.2f\" pgpgout=\"%.2f\" "
-	         "fault=\"%.2f\" majflt=\"%.2f\"/>",
+	         "fault=\"%.2f\" majflt=\"%.2f\" pgfree=\"%.2f\" "
+	         "pgscank=\"%.2f\" pgscand=\"%.2f\" pgsteal=\"%.2f\" "
+	         "vmeff-percent=\"%.2f\"/>",
 	   S_VALUE(fsj->pgpgin, fsi->pgpgin, itv),
 	   S_VALUE(fsj->pgpgout, fsi->pgpgout, itv),
 	   S_VALUE(fsj->pgfault, fsi->pgfault, itv),
-	   S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv));
+	   S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv),
+	   S_VALUE(fsj->pgfree, fsi->pgfree, itv),
+	   S_VALUE(fsj->pgscan_kswapd, fsi->pgscan_kswapd, itv),
+	   S_VALUE(fsj->pgscan_direct, fsi->pgscan_direct, itv),
+	   S_VALUE(fsj->pgsteal, fsi->pgsteal, itv),
+	   (fsi->pgscan_kswapd + fsi->pgscan_direct - fsj->pgscan_kswapd - fsj->pgscan_direct) ?
+	   SP_VALUE(fsj->pgsteal, fsi->pgsteal,
+		    fsi->pgscan_kswapd + fsi->pgscan_direct -
+		    fsj->pgscan_kswapd - fsj->pgscan_direct) : 0.0);
 
    /* memory */
    xprintf(*tab, "<memory per=\"second\" unit=\"kB\">");
@@ -1340,11 +1384,11 @@ void write_xml_stats(short curr, short *tab)
    xprintf(*tab, "<buffers>%lu</buffers>", fsi->bufkb);
    xprintf(*tab, "<cached>%lu</cached>", fsi->camkb);
    xprintf(*tab, "<frmpg>%.2f</frmpg>",
-	   S_VALUE((double) PG(fsj->frmkb), (double) PG(fsi->frmkb), itv));
+	   S_VALUE((double) KB_TO_PG(fsj->frmkb), (double) KB_TO_PG(fsi->frmkb), itv));
    xprintf(*tab, "<bufpg>%.2f</bufpg>",
-	   S_VALUE((double) PG(fsj->bufkb), (double) PG(fsi->bufkb), itv));
+	   S_VALUE((double) KB_TO_PG(fsj->bufkb), (double) KB_TO_PG(fsi->bufkb), itv));
    xprintf(*tab, "<campg>%.2f</campg>",
-	   S_VALUE((double) PG(fsj->camkb), (double) PG(fsi->camkb), itv));
+	   S_VALUE((double) KB_TO_PG(fsj->camkb), (double) KB_TO_PG(fsi->camkb), itv));
 	
    xprintf(--(*tab), "</memory>");
 
@@ -1352,20 +1396,9 @@ void write_xml_stats(short curr, short *tab)
    xprintf(*tab, "<kernel per=\"second\">");
 
    xprintf(++(*tab), "<dentunusd>%u</dentunusd>", fsi->dentry_stat);
-   xprintf(*tab, "<file-sz>%u</file-sz>", fsi->file_used);
-   xprintf(*tab, "<inode-sz>%u</inode-sz>", fsi->inode_used);
-   xprintf(*tab, "<super-sz>%u</super-sz>", fsi->super_used);
-   xprintf(*tab, "<super-sz-percent>%.2f</super-sz-percent>",
-	   fsi->super_max ?
-	   ((double) (fsi->super_used * 100)) / fsi->super_max : 0.0);
-   xprintf(*tab, "<dquot-sz>%u</dquot-sz>", fsi->dquot_used);
-   xprintf(*tab, "<dquot-sz-percent>%.2f</dquot-sz-percent>",
-	   fsi->dquot_max ?
-	   ((double) (fsi->dquot_used * 100)) / fsi->dquot_max : 0.0);
-   xprintf(*tab, "<rtsig-sz>%u</rtsig-sz>", fsi->rtsig_queued);
-   xprintf(*tab, "<rtsig-sz-percent>%.2f</rtsig-sz-percent>",
-	   fsi->rtsig_max ?
-	   ((double) (fsi->rtsig_queued * 100)) / fsi->rtsig_max : 0.0);
+   xprintf(*tab, "<file-nr>%u</file-nr>", fsi->file_used);
+   xprintf(*tab, "<inode-nr>%u</inode-nr>", fsi->inode_used);
+   xprintf(*tab, "<pty-nr>%u</pty-nr>", fsi->pty_nr);
 
    xprintf(--(*tab), "</kernel>");
 
@@ -1387,7 +1420,7 @@ void write_xml_stats(short curr, short *tab)
  * Display XML restart records
  ***************************************************************************
  */
-void write_xml_restarts(short curr, short *tab)
+void write_xml_restarts(int curr, int *tab)
 {
    char cur_time[64];
 
@@ -1401,26 +1434,60 @@ void write_xml_restarts(short curr, short *tab)
 
 /*
  ***************************************************************************
- * Print a Linux restart message (contents of a DUMMY record)
+ * Display XML COMMENT records
  ***************************************************************************
  */
-void write_dummy(short curr, int use_tm_start, int use_tm_end)
+void write_xml_comments(int curr, int *tab)
+{
+   char cur_time[64];
+   struct file_comment *file_comment;
+
+   file_comment = (struct file_comment *) &(file_stats[curr]);
+
+   /* Set timestamp for current data */
+   set_rectime(curr);
+
+   strftime(cur_time, 64, "date=\"%Y-%m-%d\" time=\"%H:%M:%S\"", &rectime);
+   xprintf(*tab, "<comment %s com=\"%s\"/>", cur_time, file_comment->comment);
+}
+
+
+/*
+ ***************************************************************************
+ * Print contents of a special (RESTART or COMMENT) record
+ ***************************************************************************
+ */
+void write_special(int curr, int use_tm_start, int use_tm_end, int rtype)
 {
    char cur_time[26];
 
    set_timestamp(curr, cur_time, 26);
 
-   /* The RESTART message must be in the interval specified by -s/-e options */
+   /* The record must be in the interval specified by -s/-e options */
    if ((use_tm_start && (datecmp(&loctime, &tm_start) < 0)) ||
        (use_tm_end && (datecmp(&loctime, &tm_end) > 0)))
       return;
 
-   if (format == S_O_PPC_OPTION)
-      printf("%s\t-1\t%ld\tLINUX-RESTART\n",
-	     file_hdr.sa_nodename, file_stats[curr].ust_time);
-   else if ((format == S_O_DB_OPTION) || (format ==S_O_DBD_OPTION))
-      printf("%s;-1;%s;LINUX-RESTART\n",
-	     file_hdr.sa_nodename, cur_time);
+   if (rtype == R_RESTART) {
+      if (format == S_O_PPC_OPTION)
+	 printf("%s\t-1\t%ld\tLINUX-RESTART\n",
+		file_hdr.sa_nodename, file_stats[curr].ust_time);
+      else if ((format == S_O_DB_OPTION) || (format ==S_O_DBD_OPTION))
+	 printf("%s;-1;%s;LINUX-RESTART\n",
+		file_hdr.sa_nodename, cur_time);
+   }
+   else if ((rtype == R_COMMENT) && DISPLAY_COMMENT(flags)) {
+      struct file_comment *file_comment;
+
+      file_comment = (struct file_comment *) &(file_stats[curr]);
+      if (format == S_O_PPC_OPTION)
+	 printf("%s\t-1\t%ld\tCOM %s\n",
+		file_hdr.sa_nodename, file_stats[curr].ust_time,
+		file_comment->comment);
+      else if ((format == S_O_DB_OPTION) || (format ==S_O_DBD_OPTION))
+	 printf("%s;-1;%s;COM %s\n",
+		file_hdr.sa_nodename, cur_time,	file_comment->comment);
+   }
 }
 
 
@@ -1433,8 +1500,9 @@ void display_file_header(char *dfile, struct file_hdr *file_hdr)
 {
    printf("File: %s (%#x)\n", dfile, file_hdr->sa_magic);
 
-   print_gal_header(localtime(&(file_hdr->sa_ust_time)), file_hdr->sa_sysname,
-		    file_hdr->sa_release, file_hdr->sa_nodename);
+   print_gal_header(localtime((const time_t *) &(file_hdr->sa_ust_time)),
+		    file_hdr->sa_sysname, file_hdr->sa_release,
+		    file_hdr->sa_nodename);
 
    printf("Activity flag: %#x\n", file_hdr->sa_actflag);
    printf("Sizeof(long): %d\n", file_hdr->sa_sizeof_long);
@@ -1451,11 +1519,11 @@ void display_file_header(char *dfile, struct file_hdr *file_hdr)
  * Display XML header and host data
  ***************************************************************************
  */
-void display_xml_header(short *tab)
+void display_xml_header(int *tab)
 {
    printf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    printf("<!DOCTYPE Configure PUBLIC \"DTD v%s sysstat //EN\"\n", XML_DTD_VERSION);
-   printf("\"http://perso.orange.fr/sebastien.godard/sysstat.dtd\">\n");
+   printf("\"http://pagesperso-orange.fr/sebastien.godard/sysstat.dtd\">\n");
 
    xprintf(*tab, "<sysstat>");
    xprintf(++(*tab), "<sysdata-version>%s</sysdata-version>", XML_DTD_VERSION);
@@ -1471,7 +1539,7 @@ void display_xml_header(short *tab)
  * Allocate structures
  ***************************************************************************
  */
-void allocate_structures(int stype)
+void allocate_structures(void)
 {
    if (file_hdr.sa_proc)
       salloc_cpu_array(st_cpu, file_hdr.sa_proc);
@@ -1521,7 +1589,7 @@ void copy_structures(int dest, int src)
  * Read varying part of the statistics from a daily data file
  ***************************************************************************
  */
-void read_extra_stats(short curr, int ifd)
+void read_extra_stats(int curr, int ifd)
 {
    if (file_hdr.sa_proc)
       sa_fread(ifd, st_cpu[curr],
@@ -1541,16 +1609,15 @@ void read_extra_stats(short curr, int ifd)
    if (file_hdr.sa_nr_disk)
       sa_fread(ifd, st_disk[curr],
 	       DISK_STATS_SIZE * file_hdr.sa_nr_disk, HARD_SIZE);
-   /* PID stats cannot be saved in file. So we don't read them */
 }
 
 
 /*
  ***************************************************************************
- * Read stats for current activity from file
+ * Read stats for current activity from file and write them
  ***************************************************************************
  */
-void read_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eosaf,
+void read_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf,
 			 unsigned int act, int *reset)
 {
    unsigned char rtype;
@@ -1575,10 +1642,17 @@ void read_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eosaf
 			file_hdr.sa_st_size, SOFT_SIZE);
       rtype = file_stats[*curr].record_type;
 	
-      if (!(*eosaf) && (rtype != R_DUMMY)) {
-	 /* Read the extra fields since it's not a DUMMY record */
+      if (!(*eosaf) && (rtype != R_RESTART) && (rtype != R_COMMENT))
+	 /* Read the extra fields since it's not a RESTART record */
 	 read_extra_stats(*curr, ifd);
 
+      if (!(*eosaf) && (rtype != R_RESTART)) {
+	
+	 if (rtype == R_COMMENT) {
+	    write_special(*curr, tm_start.use, tm_end.use, R_COMMENT);
+	    continue;
+	 }
+
 	 next = write_parsable_stats(*curr, act, *reset, cnt,
 				     tm_start.use, tm_end.use);
 	
@@ -1596,7 +1670,7 @@ void read_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eosaf
 	 *reset = FALSE;
       }
    }
-   while ((*cnt) && !(*eosaf) && (rtype != R_DUMMY));
+   while ((*cnt) && !(*eosaf) && (rtype != R_RESTART));
 
    *reset = TRUE;
 }
@@ -1609,8 +1683,7 @@ void read_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eosaf
  */
 void xml_display_loop(int ifd)
 {
-   unsigned char rtype;
-   short curr, tab = 0;
+   int curr, tab = 0, rtype;
    int eosaf = TRUE;
    off_t fpos;
 
@@ -1622,25 +1695,27 @@ void xml_display_loop(int ifd)
 
    /* Print XML header */
    display_xml_header(&tab);
-   xprintf(tab, "<statistics>");
 
    /* Process activities */
-   tab++;
+   xprintf(tab++, "<statistics>");
    do {
-      /* If this record is a DUMMY one, try to get another one */
+      /*
+       * If this record is a special (RESTART or COMMENT)  one,
+       * try to get another one.
+       */
       do {
 	 eosaf = sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE);
 	 rtype = file_stats[0].record_type;
 	
-	 if (!eosaf && (rtype != R_DUMMY)) {
+	 if (!eosaf && (rtype != R_RESTART) && (rtype != R_COMMENT)) {
 	    /*
-	     * Ok: previous record was not a DUMMY one.
+	     * Ok: previous record was not a special one.
 	     * So read now the extra fields.
 	     */
 	    read_extra_stats(0, ifd);
 	 }
       }
-      while (!eosaf && (rtype == R_DUMMY));
+      while (!eosaf && ((rtype == R_RESTART) || (rtype == R_COMMENT)));
 
       curr = 1;
 
@@ -1650,19 +1725,20 @@ void xml_display_loop(int ifd)
 			     file_hdr.sa_st_size, SOFT_SIZE);
 	    rtype = file_stats[curr].record_type;
 	
-	    if (!eosaf && (rtype != R_DUMMY)) {
-	       /* Read the extra fields since it's not a DUMMY record */
+	    if (!eosaf && (rtype != R_RESTART) && (rtype != R_COMMENT)) {
+	       /* Read the extra fields since it's not a special record */
 	       read_extra_stats(curr, ifd);
 
 	       write_xml_stats(curr, &tab);
 	       curr ^= 1;
 	    }
 	 }
-	 while (!eosaf && (rtype != R_DUMMY));
+	 while (!eosaf && (rtype != R_RESTART));
       }
 
    }
    while (!eosaf);
+   xprintf(--tab, "</statistics>");
 
    /* Rewind file */
    if (lseek(ifd, fpos, SEEK_SET) < fpos) {
@@ -1670,23 +1746,42 @@ void xml_display_loop(int ifd)
       exit(2);
    }
 
-   xprintf(--tab, "</statistics>");
-   xprintf(tab, "<restarts>");
-
-   /* Process now DUMMY entries to display restart messages */
-   tab++;
+   /* Process now RESTART entries to display restart messages */
+   xprintf(tab++, "<restarts>");
    do {
       if ((eosaf = sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE)) == 0) {
 
-	 if (file_stats[0].record_type != R_DUMMY)
+	 rtype = file_stats[0].record_type;
+	 if ((rtype != R_RESTART) && (rtype != R_COMMENT))
 	    read_extra_stats(0, ifd);
-	 else
+	 if (rtype == R_RESTART)
 	   write_xml_restarts(0, &tab);
       }
    }
    while (!eosaf);
-
    xprintf(--tab, "</restarts>");
+
+   /* Rewind file */
+   if (lseek(ifd, fpos, SEEK_SET) < fpos) {
+      perror("lseek");
+      exit(2);
+   }
+
+   /* Last, process COMMENT entries to display comments */
+   xprintf(tab++, "<comments>");
+   do {
+      if ((eosaf = sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE)) == 0) {
+
+	 rtype = file_stats[0].record_type;
+	 if ((rtype != R_RESTART) && (rtype != R_COMMENT))
+	    read_extra_stats(0, ifd);
+	 if (rtype == R_COMMENT)
+	   write_xml_comments(0, &tab);
+      }
+   }
+   while (!eosaf);
+   xprintf(--tab, "</comments>");
+
    xprintf(--tab, "</host>");
    xprintf(--tab, "</sysstat>");
 }
@@ -1699,7 +1794,7 @@ void xml_display_loop(int ifd)
  */
 void main_display_loop(int ifd)
 {
-   short curr = 1;
+   int curr = 1, rtype;
    unsigned int act;
    int eosaf = TRUE, reset = FALSE;
    long cnt = 1;
@@ -1708,24 +1803,26 @@ void main_display_loop(int ifd)
    /* Read system statistics from file */
    do {
       /*
-       * If this record is a DUMMY one, print it and (try to) get another one.
+       * If this record is a special (RESTART or COMMENT) one, print it and
+       * (try to) get another one.
        * We must be sure that we have real stats in file_stats[2].
        */
       do {
 	 if (sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE))
 	    /* End of sa data file */
 	    return;
-	
-	 if (file_stats[0].record_type == R_DUMMY)
-	    write_dummy(0, tm_start.use, tm_end.use);
+
+	 rtype = file_stats[0].record_type;
+	 if ((rtype == R_RESTART) || (rtype == R_COMMENT))
+	    write_special(0, tm_start.use, tm_end.use, rtype);
 	 else {
 	    /*
-	     Ok: previous record was not a DUMMY one. So read now the extra fields. */
+	     Ok: previous record was not a RESTART one. So read now the extra fields. */
 	    read_extra_stats(0, ifd);
 	    set_rectime(0);
 	 }
       }
-      while ((file_stats[0].record_type == R_DUMMY) ||
+      while ((rtype == R_RESTART) || (rtype == R_COMMENT) ||
 	     (tm_start.use && (datecmp(&loctime, &tm_start) < 0)) ||
 	     (tm_end.use && (datecmp(&loctime, &tm_end) >=0)));
 
@@ -1765,15 +1862,16 @@ void main_display_loop(int ifd)
 	 do {
 	    eosaf = sa_fread(ifd, &file_stats[curr],
 			     file_hdr.sa_st_size, SOFT_SIZE);
-	    if (!eosaf && (file_stats[curr].record_type != R_DUMMY))
+	    rtype = file_stats[curr].record_type;
+	    if (!eosaf && (rtype != R_RESTART) && (rtype != R_COMMENT))
 	       read_extra_stats(curr, ifd);
 	 }
-	 while (!eosaf && (file_stats[curr].record_type != R_DUMMY));
+	 while (!eosaf && (rtype != R_RESTART));
       }
 
-      /* The last record we read was a DUMMY one: print it */
-      if (!eosaf && (file_stats[curr].record_type == R_DUMMY))
-	 write_dummy(curr, tm_start.use, tm_end.use);
+      /* The last record we read was a RESTART one: Print it */
+      if (!eosaf && (file_stats[curr].record_type == R_RESTART))
+	 write_special(curr, tm_start.use, tm_end.use, R_RESTART);
    }
    while (!eosaf);
 }
@@ -1798,7 +1896,7 @@ void read_stats_from_file(char dfile[])
    }
 
    /* Perform required allocations */
-   allocate_structures(USE_SA_FILE);
+   allocate_structures();
 
    set_hdr_rectime(flags, &rectime, &file_hdr);
 
@@ -1985,6 +2083,9 @@ int main(int argc, char **argv)
    if (!dfile[0])
       set_default_file(&rectime, dfile);
 
+   if (tm_start.use && tm_end.use && (tm_end.tm_hour < tm_start.tm_hour))
+      tm_end.tm_hour += 24;
+
    /*
     * Display all the contents of the daily data file if the count parameter
     * was not set on the command line.
