@@ -1,6 +1,6 @@
 /*
  * sar: report system activity
- * (C) 1999-2006 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1999-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -26,8 +26,10 @@
 #include <time.h>
 #include <errno.h>
 
+#include "version.h"
 #include "sa.h"
 #include "common.h"
+#include "ioconf.h"
 
 
 #ifdef USE_NLS
@@ -38,6 +40,9 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)sysstat-" VERSION ": " __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
+
 long interval = -1, count = 0;
 unsigned int sar_actflag = 0;
 unsigned int flags = 0;
@@ -46,22 +51,19 @@ unsigned char cpu_bitmap[(NR_CPUS / 8) + 1];
 
 struct stats_sum asum;
 struct file_hdr file_hdr;
-struct file_stats file_stats[DIM];
-struct stats_one_cpu *st_cpu[DIM] = {NULL, NULL, NULL};
-struct stats_serial *st_serial[DIM] = {NULL, NULL, NULL};
-struct stats_irq_cpu *st_irq_cpu[DIM] = {NULL, NULL, NULL};
-struct stats_net_dev *st_net_dev[DIM] = {NULL, NULL, NULL};
-struct disk_stats *st_disk[DIM] = {NULL, NULL, NULL};
+struct file_stats file_stats[3];
+struct stats_one_cpu *st_cpu[3] = {NULL, NULL, NULL};
+struct stats_serial *st_serial[3] = {NULL, NULL, NULL};
+struct stats_irq_cpu *st_irq_cpu[3] = {NULL, NULL, NULL};
+struct stats_net_dev *st_net_dev[3] = {NULL, NULL, NULL};
+struct disk_stats *st_disk[3] = {NULL, NULL, NULL};
 
 /* Array members of common types are always packed */
-unsigned int interrupts[DIM][NR_IRQS];
-/* Structures are aligned but also padded. Thus array members are packed */
-struct pid_stats *pid_stats[DIM][MAX_PID_NR];
+unsigned int interrupts[3][NR_IRQS];
 
 struct tm rectime;
 /* Contain the date specified by -s and -e options */
 struct tstamp tm_start, tm_end;
-int pid_nr = 0;
 char *args[MAX_ARGV_NR];
 
 /*
@@ -73,11 +75,10 @@ void usage(char *progname)
 {
    fprintf(stderr, _("Usage: %s [ options... ] [ <interval> [ <count> ] ]\n"
 	           "Options are:\n"
-	           "[ -A ] [ -b ] [ -B ] [ -c ] [ -d ] [ -i <interval> ] [ -p ] [ -q ]\n"
+	           "[ -A ] [ -b ] [ -B ] [ -c ] [ -C ] [ -d ] [ -i <interval> ] [ -p ] [ -q ]\n"
 		   "[ -r ] [ -R ] [ -t ] [ -u ] [ -v ] [ -V ] [ -w ] [ -W ] [ -y ]\n"
 		   "[ -I { <irq> | SUM | ALL | XALL } ] [ -P { <cpu> | ALL } ]\n"
 		   "[ -n { DEV | EDEV | NFS | NFSD | SOCK | ALL } ]\n"
-		   "[ -x { <pid> | SELF | ALL } ] [ -X { <pid> | SELF | ALL } ]\n"
 	           "[ -o [ <filename> ] | -f [ <filename> ] ]\n"
 		   "[ -s [ <hh:mm:ss> ] ] [ -e [ <hh:mm:ss> ] ]\n"),
 	   progname);
@@ -92,13 +93,8 @@ void usage(char *progname)
  */
 void init_all_stats(void)
 {
-   int i;
-
    init_stats(file_stats, interrupts);
    memset(&asum, 0, STATS_SUM_SIZE);
-
-   for (i = 0; i < DIM; i++)
-      pid_stats[i][0] = NULL;
 }
 
 
@@ -117,46 +113,15 @@ void salloc(int i, char *ltemp)
 }
 
 
-/*
- ***************************************************************************
- * Allocate pid_stats structures
- ***************************************************************************
- */
-void salloc_pid(int nr_pid)
-{
-   int i, pid;
-
-   for (i = 0; i < DIM; i++) {
-      if (pid_stats[i][0])
-	 free(pid_stats[i][0]);
-      if ((pid_stats[i][0] = (struct pid_stats *) malloc(PID_STATS_SIZE * nr_pid)) == NULL) {
-	 perror("malloc");
-	 exit(4);
-      }
-
-      memset(pid_stats[i][0], 0, PID_STATS_SIZE * nr_pid);
-
-      for (pid = 1; pid < nr_pid; pid++)
-	 /* Structures are aligned but also padded. Thus array members are packed */
-	 pid_stats[i][pid] = pid_stats[i][0] + pid;	/* Assume nr_pids <= MAX_PID_NR */
-   }
-}
-
-
 /*
  ***************************************************************************
  * Allocate structures
  ***************************************************************************
  */
-void allocate_structures(int stype)
+void allocate_structures(void)
 {
    if (file_hdr.sa_proc)
       salloc_cpu_array(st_cpu, file_hdr.sa_proc);
-   if ((stype == USE_SADC) &&
-       (GET_PID(sar_actflag) || GET_CPID(sar_actflag))) {
-      pid_nr = file_hdr.sa_nr_pid;
-      salloc_pid(pid_nr);
-   }
    if (file_hdr.sa_serial)
       salloc_serial_array(st_serial, file_hdr.sa_serial);
    if (file_hdr.sa_irqcpu)
@@ -224,7 +189,7 @@ void prep_smp_option(unsigned int cpu_nr)
  * structure.
  ***************************************************************************
 */
-void set_rectime(short curr)
+void set_rectime(int curr)
 {
    struct tm *ltm;
 
@@ -236,7 +201,7 @@ void set_rectime(short curr)
       rectime.tm_sec  = file_stats[curr].second;
    }
    else {
-      ltm = localtime(&file_stats[curr].ust_time);
+      ltm = localtime((const time_t *) &file_stats[curr].ust_time);
       rectime = *ltm;
    }
 }
@@ -267,12 +232,39 @@ int count_bits(void *ptr, int size)
 }
 
 
+/*
+ ***************************************************************************
+ * Determine if a stat line header has to be displayed
+ ***************************************************************************
+*/
+void check_line_hdr(int *dis_hdr)
+{
+   int nr_opt;
+
+   /* Get number of options entered on the command line */
+   nr_opt = count_bits(&sar_actflag, sizeof(unsigned int));
+
+   if ((nr_opt > 1) ||
+       ((GET_NET_DEV(sar_actflag) || GET_NET_EDEV(sar_actflag)) && (file_hdr.sa_iface > 1)) ||
+       (GET_DISK(sar_actflag) && (file_hdr.sa_nr_disk > 1)))
+      *dis_hdr = 1;
+   else if (GET_ONE_IRQ(sar_actflag)) {
+      if (count_bits(irq_bitmap, sizeof(irq_bitmap)) > 1)
+	 *dis_hdr = 1;
+   }
+   else if ((GET_IRQ(sar_actflag) || GET_CPU(sar_actflag)) && WANT_PER_PROC(flags)) {
+      if (count_bits(cpu_bitmap, sizeof(cpu_bitmap)) > 1)
+	 *dis_hdr = 1;
+   }
+}
+
+
 /*
  ***************************************************************************
  * Set timestamp string
  ***************************************************************************
 */
-void set_timestamp(short curr, char *cur_time, int len)
+void set_timestamp(int curr, char *cur_time, int len)
 {
    set_rectime(curr);
 
@@ -286,7 +278,7 @@ void set_timestamp(short curr, char *cur_time, int len)
  * Core function used to display statistics
  ***************************************************************************
  */
-void write_stats_core(short prev, short curr, short dis, char *prev_string,
+void write_stats_core(int prev, int curr, int dis, char *prev_string,
 		      char *curr_string, unsigned int act,
 		      unsigned long long itv, unsigned long long g_itv,
 		      int disp_avg, int want_since_boot)
@@ -296,14 +288,8 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
       *fsi = &file_stats[curr],
       *fsj = &file_stats[prev];
 
-   /*
-    * Under very special circumstances, STDOUT may become unavailable,
-    * This is what we try to guess here
-    */
-   if (write(STDOUT_FILENO, "", 0) == -1) {
-      perror("stdout");
-      exit(6);
-   }
+   /* Test stdout */
+   TEST_STDOUT(STDOUT_FILENO);
 
    /* Print number of processes created per second */
    if (GET_PROC(act)) {
@@ -412,14 +398,24 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
    /* Print paging statistics */
    if (GET_PAGE(act)) {
       if (dis)
-	 printf("\n%-11s  pgpgin/s pgpgout/s   fault/s  majflt/s\n",
+	 printf("\n%-11s  pgpgin/s pgpgout/s   fault/s  majflt/s  pgfree/s"
+		" pgscank/s pgscand/s pgsteal/s    %%vmeff\n",
 		prev_string);
 
-      printf("%-11s %9.2f %9.2f %9.2f %9.2f\n", curr_string,
+      printf("%-11s %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
+	     curr_string,
 	     S_VALUE(fsj->pgpgin, fsi->pgpgin, itv),
 	     S_VALUE(fsj->pgpgout, fsi->pgpgout, itv),
 	     S_VALUE(fsj->pgfault, fsi->pgfault, itv),
-	     S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv));
+	     S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv),
+	     S_VALUE(fsj->pgfree, fsi->pgfree, itv),
+	     S_VALUE(fsj->pgscan_kswapd, fsi->pgscan_kswapd, itv),
+	     S_VALUE(fsj->pgscan_direct, fsi->pgscan_direct, itv),
+	     S_VALUE(fsj->pgsteal, fsi->pgsteal, itv),
+	     (fsi->pgscan_kswapd + fsi->pgscan_direct - fsj->pgscan_kswapd - fsj->pgscan_direct) ?
+	     SP_VALUE(fsj->pgsteal, fsi->pgsteal,
+		      fsi->pgscan_kswapd + fsi->pgscan_direct -
+		      fsj->pgscan_kswapd - fsj->pgscan_direct) : 0.0);
    }
 
    /* Print number of swap pages brought in and out */
@@ -452,62 +448,9 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
 	 printf("\n%-11s   frmpg/s   bufpg/s   campg/s\n", prev_string);
 
       printf("%-11s %9.2f %9.2f %9.2f\n", curr_string,
-	     S_VALUE((double) PG(fsj->frmkb), (double) PG(fsi->frmkb), itv),
-	     S_VALUE((double) PG(fsj->bufkb), (double) PG(fsi->bufkb), itv),
-	     S_VALUE((double) PG(fsj->camkb), (double) PG(fsi->camkb), itv));
-   }
-
-   /* Print per-process statistics */
-   if (GET_PID(act)) {
-      if (dis) {
-	 printf("\n%-11s       PID  minflt/s  majflt/s     %%user   %%system   nswap/s",
-		prev_string);
-	 if (!disp_avg)
-	    printf("   CPU\n");
-	 else
-	    printf("\n");
-      }
-
-      for (i = 0; i < pid_nr; i++) {
-	 if (!pid_stats[curr][i]->pid || !(pid_stats[curr][i]->flag & X_PID_SET))
-	    continue;
-
- 	 printf("%-11s %9ld", curr_string, pid_stats[curr][i]->pid);
-
-	 printf(" %9.2f %9.2f    %6.2f    %6.2f %9.2f",
-		S_VALUE(pid_stats[prev][i]->minflt, pid_stats[curr][i]->minflt, itv),
-		S_VALUE(pid_stats[prev][i]->majflt, pid_stats[curr][i]->majflt, itv),
-		SP_VALUE(pid_stats[prev][i]->utime, pid_stats[curr][i]->utime, itv),
-		SP_VALUE(pid_stats[prev][i]->stime, pid_stats[curr][i]->stime, itv),
-		S_VALUE((long) pid_stats[prev][i]->nswap,
-			(long) pid_stats[curr][i]->nswap, itv));
-
-	 if (!disp_avg)
-	    printf("   %3d\n", pid_stats[curr][i]->processor);
-	 else
-	    printf("\n");
-      }
-   }
-
-   /* Print statistics about children of a given process */
-   if (GET_CPID(act)) {
-      if (dis)
-	 printf("\n%-11s      PPID cminflt/s cmajflt/s    %%cuser  %%csystem  cnswap/s\n",
-		prev_string);
-
-      for (i = 0; i < pid_nr; i++) {
-	 if (!pid_stats[curr][i]->pid || !(pid_stats[curr][i]->flag & X_PPID_SET))
-	    continue;
-	 printf("%-11s %9ld", curr_string, pid_stats[curr][i]->pid);
-
-	 printf(" %9.2f %9.2f    %6.2f    %6.2f %9.2f\n",
-		S_VALUE(pid_stats[prev][i]->cminflt, pid_stats[curr][i]->cminflt, itv),
-		S_VALUE(pid_stats[prev][i]->cmajflt, pid_stats[curr][i]->cmajflt, itv),
-		SP_VALUE(pid_stats[prev][i]->cutime,  pid_stats[curr][i]->cutime,  itv),
-		SP_VALUE(pid_stats[prev][i]->cstime,  pid_stats[curr][i]->cstime,  itv),
-		S_VALUE((long) pid_stats[prev][i]->cnswap,
-			(long) pid_stats[curr][i]->cnswap, itv));
-      }
+	     S_VALUE((double) KB_TO_PG(fsj->frmkb), (double) KB_TO_PG(fsi->frmkb), itv),
+	     S_VALUE((double) KB_TO_PG(fsj->bufkb), (double) KB_TO_PG(fsi->bufkb), itv),
+	     S_VALUE((double) KB_TO_PG(fsj->camkb), (double) KB_TO_PG(fsi->camkb), itv));
    }
 
    /* Print TTY statistics (serial lines) */
@@ -624,7 +567,7 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
 	 *sndj;
 
       if (dis)
-	 printf("\n%-11s     IFACE   rxpck/s   txpck/s   rxbyt/s   txbyt/s   "
+	 printf("\n%-11s     IFACE   rxpck/s   txpck/s    rxkB/s    txkB/s   "
 		"rxcmp/s   txcmp/s  rxmcst/s\n",
 		prev_string);
 
@@ -639,8 +582,8 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
 	 printf(" %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
 		S_VALUE(sndj->rx_packets, sndi->rx_packets, itv),
 		S_VALUE(sndj->tx_packets, sndi->tx_packets, itv),
-		S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv),
-		S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv),
+		S_VALUE(sndj->rx_bytes, sndi->rx_bytes, itv) / 1024,
+		S_VALUE(sndj->tx_bytes, sndi->tx_bytes, itv) / 1024,
 		S_VALUE(sndj->rx_compressed, sndi->rx_compressed, itv),
 		S_VALUE(sndj->tx_compressed, sndi->tx_compressed, itv),
 		S_VALUE(sndj->multicast, sndi->multicast, itv));
@@ -686,6 +629,7 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
       struct disk_stats
 	 *sdi = st_disk[curr],
 	 *sdj;
+      char *dev_name;
 
       if (dis)
 	 printf("\n%-11s       DEV       tps  rd_sec/s  wr_sec/s  avgrq-sz  "
@@ -710,10 +654,18 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
 	    ((sdi->rd_sect - sdj->rd_sect) + (sdi->wr_sect - sdj->wr_sect)) /
 	    ((double) (sdi->nr_ios - sdj->nr_ios)) : 0.0;
 
+	 dev_name = NULL;
+
+	 if ((USE_PRETTY_OPTION(flags)) && (sdi->major == DEVMAP_MAJOR))
+	    dev_name = transform_devmapname(sdi->major, sdi->minor);
+	
+	 if (!dev_name)
+	    dev_name = get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags));
+	
 	 printf("%-11s %9s %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
 		curr_string,
 		/* Confusion possible here between index and minor numbers */
-		get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags)),
+		dev_name,
 		S_VALUE(sdj->nr_ios, sdi->nr_ios,  itv),
 		ll_s_value(sdj->rd_sect, sdi->rd_sect, itv),
 		ll_s_value(sdj->wr_sect, sdi->wr_sect, itv),
@@ -770,7 +722,7 @@ void write_stats_core(short prev, short curr, short dis, char *prev_string,
  * Print statistics average
  ***************************************************************************
  */
-void write_stats_avg(int curr, short dis, unsigned int act, int read_from_file)
+void write_stats_avg(int curr, int dis, unsigned int act, int read_from_file)
 {
    unsigned long long itv, g_itv;
    char string[16];
@@ -778,21 +730,13 @@ void write_stats_avg(int curr, short dis, unsigned int act, int read_from_file)
      *fsi = &file_stats[curr];
 
    /* Interval value in jiffies */
-   g_itv = (file_stats[curr].uptime - file_stats[2].uptime) & 0xffffffff;
+   g_itv = get_interval(file_stats[2].uptime, file_stats[curr].uptime);
+
    if (file_hdr.sa_proc)
-      itv = (file_stats[curr].uptime0 - file_stats[2].uptime0) & 0xffffffff;
+      itv = get_interval(file_stats[2].uptime0, file_stats[curr].uptime0);
    else
       itv = g_itv;
 
-   if (!itv || !g_itv) { /* Should no longer happen with recent versions */
-      /*
-       * Aiee: null interval... This should only happen when reading stats
-       * from a system activity file with a far too big interval value...
-       */
-      fprintf(stderr, _("Please give a smaller interval value\n"));
-      exit(1);
-   }
-
    strcpy(string, _("Average:"));
    write_stats_core(2, curr, dis, string, string, act, itv, g_itv, TRUE, FALSE);
 
@@ -821,40 +765,29 @@ void write_stats_avg(int curr, short dis, unsigned int act, int read_from_file)
 
    if (GET_KTABLES(act)) {
       if (dis)
-	 printf("\n%-11s dentunusd   file-sz  inode-sz  super-sz %%super-sz  "
-		"dquot-sz %%dquot-sz  rtsig-sz %%rtsig-sz\n",
+	 printf("\n%-11s dentunusd   file-nr  inode-nr    pty-nr\n",
 		string);
 
-      printf("%-11s %9.0f %9.0f %9.0f %9.0f    %6.2f %9.0f    %6.2f %9.0f    %6.2f\n",
+      printf("%-11s %9.0f %9.0f %9.0f %9.0f\n",
 	     string,
 	     (double) asum.dentry_stat / asum.count,
 	     (double) asum.file_used / asum.count,
 	     (double) asum.inode_used / asum.count,
-	     (double) asum.super_used / asum.count,
-	     fsi->super_max ?
-	     ((double) ((asum.super_used / asum.count) * 100)) / fsi->super_max
-	     : 0.0,
-	     (double) asum.dquot_used / asum.count,
-	     fsi->dquot_max ?
-	     ((double) ((asum.dquot_used / asum.count) * 100)) / fsi->dquot_max
-	     : 0.0,
-	     (double) asum.rtsig_queued / asum.count,
-	     fsi->rtsig_max ?
-	     ((double) ((asum.rtsig_queued / asum.count) * 100)) / fsi->rtsig_max
-	     : 0.0);
+	     (double) asum.pty_nr / asum.count);
    }
 
    if (GET_NET_SOCK(act)) {
       if (dis)
-	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag\n",
+	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag    tcp-tw\n",
 		string);
 
-      printf("%-11s %9.0f %9.0f %9.0f %9.0f %9.0f\n", string,
+      printf("%-11s %9.0f %9.0f %9.0f %9.0f %9.0f %9.0f\n", string,
 	     (double) asum.sock_inuse / asum.count,
 	     (double) asum.tcp_inuse / asum.count,
 	     (double) asum.udp_inuse / asum.count,
 	     (double) asum.raw_inuse / asum.count,
-	     (double) asum.frag_inuse / asum.count);
+	     (double) asum.frag_inuse / asum.count,
+	     (double) asum.tcp_tw / asum.count);
    }
 
    if (GET_QUEUE(act)) {
@@ -881,7 +814,7 @@ void write_stats_avg(int curr, short dis, unsigned int act, int read_from_file)
  * Print system statistics
  ***************************************************************************
  */
-int write_stats(short curr, short dis, unsigned int act, int read_from_file,
+int write_stats(int curr, int dis, unsigned int act, int read_from_file,
 		long *cnt, int use_tm_start, int use_tm_end, int reset,
 		int want_since_boot)
 {
@@ -889,11 +822,12 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
    unsigned long long itv, g_itv;
    struct file_stats
       *fsi = &file_stats[curr];
+   static int cross_day = 0;
 
    /* Check time (1) */
    if (read_from_file) {
-      if (!next_slice(file_stats[2].uptime, file_stats[curr].uptime,
-		      &file_hdr, reset, interval))
+      if (!next_slice(file_stats[2].uptime0, file_stats[curr].uptime0,
+		      reset, interval))
 	 /* Not close enough to desired interval */
 	 return 0;
    }
@@ -903,6 +837,20 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
    /* Set current timestamp */
    set_timestamp(curr, cur_time[curr], 16);
 
+   /* Check if we are beginning a new day */
+   if (use_tm_start && file_stats[!curr].ust_time &&
+       (file_stats[curr].ust_time > file_stats[!curr].ust_time) &&
+       (file_stats[curr].hour < file_stats[!curr].hour))
+      cross_day = 1;
+
+   if (cross_day)
+      /*
+       * This is necessary if we want to properly handle something like:
+       * sar -s time_start -e time_end with
+       * time_start(day D) > time_end(day D+1)
+       */
+      rectime.tm_hour +=24;
+
    /* Check time (2) */
    if (use_tm_start && (datecmp(&rectime, &tm_start) < 0))
      /* it's too soon... */
@@ -947,7 +895,6 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
 
       /*
        * Will be used to compute the average.
-       * Note: overflow unlikely to happen but not impossible...
        * We assume that the total amount of memory installed can not vary
        * during the interval given on the command line, whereas the total
        * amount of swap space may.
@@ -963,24 +910,15 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
    /* Print values of some kernel tables */
    if (GET_KTABLES(act)) {
       if (dis)
-	 printf("\n%-11s dentunusd   file-sz  inode-sz  super-sz %%super-sz  "
-		"dquot-sz %%dquot-sz  rtsig-sz %%rtsig-sz\n",
+	 printf("\n%-11s dentunusd   file-nr  inode-nr    pty-nr\n",
 		cur_time[!curr]);
 
-      printf("%-11s %9u %9u %9u %9u    %6.2f %9u    %6.2f %9u    %6.2f\n",
+      printf("%-11s %9u %9u %9u %9u\n",
 	     cur_time[curr],
 	     fsi->dentry_stat,
 	     fsi->file_used,
 	     fsi->inode_used,
-	     fsi->super_used,
-	     fsi->super_max ?
-	     ((double) (fsi->super_used * 100)) / fsi->super_max : 0.0,
-	     fsi->dquot_used,
-	     fsi->dquot_max ?
-	     ((double) (fsi->dquot_used * 100)) / fsi->dquot_max : 0.0,
-	     fsi->rtsig_queued,
-	     fsi->rtsig_max ?
-	     ((double) (fsi->rtsig_queued * 100)) / fsi->rtsig_max : 0.0);
+	     fsi->pty_nr);
 
       /*
        * Will be used to compute the average.
@@ -990,20 +928,18 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
       asum.dentry_stat += fsi->dentry_stat;
       asum.file_used += fsi->file_used;
       asum.inode_used += fsi->inode_used;
-      asum.super_used += fsi->super_used;
-      asum.dquot_used += fsi->dquot_used;
-      asum.rtsig_queued += fsi->rtsig_queued;
+      asum.pty_nr += fsi->pty_nr;
    }
 
    /* Print number of sockets in use */
    if (GET_NET_SOCK(act)) {
       if (dis)
-	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag\n",
+	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag    tcp-tw\n",
 		cur_time[!curr]);
 
-      printf("%-11s %9u %9u %9u %9u %9u\n", cur_time[curr],
+      printf("%-11s %9u %9u %9u %9u %9u %9u\n", cur_time[curr],
 	     fsi->sock_inuse, fsi->tcp_inuse, fsi->udp_inuse,
-	     fsi->raw_inuse, fsi->frag_inuse);
+	     fsi->raw_inuse, fsi->frag_inuse, fsi->tcp_tw);
 
       /* Will be used to compute the average */
       asum.sock_inuse += fsi->sock_inuse;
@@ -1011,6 +947,7 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
       asum.udp_inuse += fsi->udp_inuse;
       asum.raw_inuse += fsi->raw_inuse;
       asum.frag_inuse += fsi->frag_inuse;
+      asum.tcp_tw += fsi->tcp_tw;
    }
 
    /* Print load averages and queue length */
@@ -1042,7 +979,7 @@ int write_stats(short curr, short dis, unsigned int act, int read_from_file,
  * Display stats since system startup
  ***************************************************************************
  */
-void write_stats_startup(short curr)
+void write_stats_startup(int curr)
 {
    /* Set to 0 previous structures corresponding to boot time */
    memset(&file_stats[!curr], 0, FILE_STATS_SIZE);
@@ -1054,8 +991,6 @@ void write_stats_startup(short curr)
    if (file_hdr.sa_proc)
       memset(st_cpu[!curr], 0, STATS_ONE_CPU_SIZE * file_hdr.sa_proc);
    memset(interrupts[!curr], 0, STATS_ONE_IRQ_SIZE);
-   if (pid_nr)
-      memset (pid_stats[!curr][0], 0, PID_STATS_SIZE * pid_nr);
    if (file_hdr.sa_serial)
       memset(st_serial[!curr], 0, STATS_SERIAL_SIZE * file_hdr.sa_serial);
    if (file_hdr.sa_irqcpu)
@@ -1101,21 +1036,34 @@ int sa_read(void *buffer, int size)
 
 /*
  ***************************************************************************
- * Print a Linux restart message (contents of a DUMMY record)
+ * Print a Linux restart message (contents of a RESTART record) or a
+ * comment (contents of a COMMENT record).
  ***************************************************************************
  */
-void write_dummy(short curr, int use_tm_start, int use_tm_end)
+int write_special(int curr, int use_tm_start, int use_tm_end, int rtype)
 {
    char cur_time[26];
 
    set_timestamp(curr, cur_time, 26);
 
-   /* The RESTART message must be in the interval specified by -s/-e options */
+   /* The record must be in the interval specified by -s/-e options */
    if ((use_tm_start && (datecmp(&rectime, &tm_start) < 0)) ||
        (use_tm_end && (datecmp(&rectime, &tm_end) > 0)))
-      return;
+      return 0;
 
-   printf("\n%-11s       LINUX RESTART\n", cur_time);
+   if (rtype == R_RESTART) {
+      printf("\n%-11s       LINUX RESTART\n", cur_time);
+      return 1;
+   }
+   else if ((rtype == R_COMMENT) && DISPLAY_COMMENT(flags)) {
+      struct file_comment *file_comment;
+
+      file_comment = (struct file_comment *) &(file_stats[curr]);
+      printf("%-11s  COM %s\n", cur_time, file_comment->comment);
+      return 1;
+   }
+
+   return 0;
 }
 
 
@@ -1124,7 +1072,7 @@ void write_dummy(short curr, int use_tm_start, int use_tm_end)
  * Move structures data
  ***************************************************************************
  */
-void copy_structures(int dest, int src, int stype)
+void copy_structures(int dest, int src)
 {
    memcpy(&file_stats[dest], &file_stats[src], FILE_STATS_SIZE);
    if (file_hdr.sa_proc)
@@ -1133,9 +1081,6 @@ void copy_structures(int dest, int src, int stype)
    if (GET_ONE_IRQ(file_hdr.sa_actflag))
       memcpy(interrupts[dest], interrupts[src],
 	     STATS_ONE_IRQ_SIZE);
-   if ((stype == USE_SADC) && (pid_nr))
-      memcpy(pid_stats[dest][0], pid_stats[src][0],
-	     PID_STATS_SIZE * pid_nr);
    if (file_hdr.sa_serial)
       memcpy(st_serial[dest], st_serial[src],
 	     STATS_SERIAL_SIZE * file_hdr.sa_serial);
@@ -1156,7 +1101,7 @@ void copy_structures(int dest, int src, int stype)
  * Read varying part of the statistics from a daily data file
  ***************************************************************************
  */
-void read_extra_stats(short curr, int ifd)
+void read_extra_stats(int curr, int ifd)
 {
    if (file_hdr.sa_proc)
       sa_fread(ifd, st_cpu[curr],
@@ -1176,7 +1121,6 @@ void read_extra_stats(short curr, int ifd)
    if (file_hdr.sa_nr_disk)
       sa_fread(ifd, st_disk[curr],
 	       DISK_STATS_SIZE * file_hdr.sa_nr_disk, HARD_SIZE);
-   /* PID stats cannot be saved in file. So we don't read them */
 }
 
 
@@ -1185,7 +1129,7 @@ void read_extra_stats(short curr, int ifd)
  * Read a bunch of statistics sent by the data collector (sadc)
  ***************************************************************************
  */
-void read_stat_bunch(short curr)
+void read_stat_bunch(int curr)
 {
    if (sa_read(&file_stats[curr], file_hdr.sa_st_size))
       exit(0);
@@ -1195,9 +1139,6 @@ void read_stat_bunch(short curr)
    if (GET_ONE_IRQ(file_hdr.sa_actflag) &&
        sa_read(interrupts[curr], STATS_ONE_IRQ_SIZE))
       exit(0);
-   if (pid_nr &&
-       sa_read(pid_stats[curr][0], PID_STATS_SIZE * pid_nr))
-      exit(0);
    if (file_hdr.sa_serial &&
        sa_read(st_serial[curr], STATS_SERIAL_SIZE * file_hdr.sa_serial))
       exit(0);
@@ -1218,10 +1159,10 @@ void read_stat_bunch(short curr)
  * Read stats for current activity from file and display them.
  ***************************************************************************
  */
-void handle_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eosaf,
+void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf,
 			   int rows, unsigned int act, int *reset, int nr_cpu, int nr_irq)
 {
-   short dis = 1;
+   int dis = 1;
    unsigned long lines;
    unsigned char rtype;
    int davg, next, inc = 1;
@@ -1235,7 +1176,7 @@ void handle_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eos
     * Restore the first stats collected.
     * Used to compute the rate displayed on the first line.
     */
-   copy_structures(!(*curr), 2, USE_SA_FILE);
+   copy_structures(!(*curr), 2);
 	
    lines = 0;
    davg  = 0;
@@ -1261,8 +1202,8 @@ void handle_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eos
 			file_hdr.sa_st_size, SOFT_SIZE);
       rtype = file_stats[*curr].record_type;
 	
-      if (!(*eosaf) && (rtype != R_DUMMY))
-	 /* Read the extra fields since it's not a DUMMY record */
+      if (!(*eosaf) && (rtype != R_RESTART) && (rtype != R_COMMENT))
+	 /* Read the extra fields since it's not a special record */
 	 read_extra_stats(*curr, ifd);
 
       if ((lines >= rows) || !lines) {
@@ -1272,9 +1213,16 @@ void handle_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eos
       else
 	 dis = 0;
 
-      lines += inc;
-
-      if (!(*eosaf) && (rtype != R_DUMMY)) {
+      if (!(*eosaf) && (rtype != R_RESTART)) {
+	
+	 if (rtype == R_COMMENT) {
+	    /* Display comment */
+	    next = write_special(*curr, tm_start.use, tm_end.use, R_COMMENT);
+	    if (next)
+	       /* A line of comment was actually displayed */
+	       lines++;
+	    continue;
+	 }
 
 	 /* next is set to 1 when we were close enough to desired interval */
 	 next = write_stats(*curr, dis, act, USE_SA_FILE, cnt,
@@ -1284,13 +1232,12 @@ void handle_curr_act_stats(int ifd, off_t fpos, short *curr, long *cnt, int *eos
 	 if (next) {
 	    davg++;
 	    *curr ^=1;
+	    lines += inc;
 	 }
-	 else
-	    lines -= inc;;
-	 *reset = 0;
+	 *reset = FALSE;
       }
    }
-   while ((*cnt) && !(*eosaf) && (rtype != R_DUMMY));
+   while ((*cnt) && !(*eosaf) && (rtype != R_RESTART));
 
    if (davg)
       write_stats_avg(!(*curr), dis, act, USE_SA_FILE);
@@ -1325,9 +1272,9 @@ void read_header_data(void)
  */
 void read_stats_from_file(char from_file[])
 {
-   short curr = 1;
+   int curr = 1;
    unsigned int act;
-   int ifd, nr_cpu, nr_irq;
+   int ifd, nr_cpu, nr_irq, rtype;
    int rows, eosaf = TRUE, reset = FALSE;
    long cnt = 1;
    off_t fpos;
@@ -1345,7 +1292,7 @@ void read_stats_from_file(char from_file[])
       nr_cpu = count_bits(cpu_bitmap, sizeof(cpu_bitmap));
 
    /* Perform required allocations */
-   allocate_structures(USE_SA_FILE);
+   allocate_structures();
 
    /* Print report header */
    print_report_hdr(flags, &rectime, &file_hdr);
@@ -1353,31 +1300,33 @@ void read_stats_from_file(char from_file[])
    /* Read system statistics from file */
    do {
       /*
-       * If this record is a DUMMY one, print it and (try to) get another one.
+       * If this record is a special (RESTART or COMMENT) one, print it and
+       * (try to) get another one.
        * We must be sure that we have real stats in file_stats[2].
        */
       do {
 	 if (sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE))
 	    /* End of sa data file */
 	    return;
-	
-	 if (file_stats[0].record_type == R_DUMMY)
-	    write_dummy(0, tm_start.use, tm_end.use);
+
+	 rtype = file_stats[0].record_type;
+	 if ((rtype == R_RESTART) || (rtype == R_COMMENT))
+	    write_special(0, tm_start.use, tm_end.use, rtype);
 	 else {
 	    /*
-	     * Ok: previous record was not a DUMMY one.
+	     * Ok: previous record was not a special one.
 	     * So read now the extra fields.
 	     */
 	    read_extra_stats(0, ifd);
 	    set_rectime(0);
 	 }
       }
-      while ((file_stats[0].record_type == R_DUMMY) ||
+      while ((rtype == R_RESTART) || (rtype == R_COMMENT) ||
 	     (tm_start.use && (datecmp(&rectime, &tm_start) < 0)) ||
 	     (tm_end.use && (datecmp(&rectime, &tm_end) >=0)));
 
       /* Save the first stats collected. Will be used to compute the average */
-      copy_structures(2, 0, USE_SA_FILE);
+      copy_structures(2, 0);
 
       reset = TRUE;	/* Set flag to reset last_uptime variable */
 
@@ -1415,15 +1364,16 @@ void read_stats_from_file(char from_file[])
 	 do {
 	    eosaf = sa_fread(ifd, &file_stats[curr],
 			     file_hdr.sa_st_size, SOFT_SIZE);
-	    if (!eosaf && (file_stats[curr].record_type != R_DUMMY))
+	    rtype = file_stats[curr].record_type;
+	    if (!eosaf && (rtype != R_RESTART) && (rtype != R_COMMENT))
 	       read_extra_stats(curr, ifd);
 	 }
-	 while (!eosaf && (file_stats[curr].record_type != R_DUMMY));
+	 while (!eosaf && (rtype != R_RESTART));
       }
 
-      /* The last record we read was a DUMMY one: print it */
-      if (!eosaf && (file_stats[curr].record_type == R_DUMMY))
-	 write_dummy(curr, tm_start.use, tm_end.use);
+      /* The last record we read was a RESTART one: Print it */
+      if (!eosaf && (file_stats[curr].record_type == R_RESTART))
+	 write_special(curr, tm_start.use, tm_end.use, R_RESTART);
    }
    while (!eosaf);
 
@@ -1438,10 +1388,10 @@ void read_stats_from_file(char from_file[])
  */
 void read_stats(void)
 {
-   short curr = 1, dis = 1;
+   int curr = 1, dis = 1;
    unsigned long lines = 0;
    unsigned int rows = 23;
-   int nr_opt, dis_hdr = 0;
+   int dis_hdr = 0;
 
    /* Read stats header */
    read_header_data();
@@ -1470,28 +1420,8 @@ void read_stats(void)
       exit(1);
    }
 
-   /* Get number of options entered on the command line */
-   nr_opt = count_bits(&sar_actflag, sizeof(unsigned int));
-
-   /* Determine if stats line header has to be displayed */
-   if ((nr_opt > 1) ||
-       ((GET_NET_DEV(sar_actflag) || GET_NET_EDEV(sar_actflag)) && (file_hdr.sa_iface > 1)) ||
-       (GET_DISK(sar_actflag) && (file_hdr.sa_nr_disk > 1)))
-      dis_hdr = 1;
-   else if (GET_ONE_IRQ(sar_actflag)) {
-      int nr_irq;
-	
-      nr_irq = count_bits(irq_bitmap, sizeof(irq_bitmap));
-      if (nr_irq > 1)
-	 dis_hdr = 1;
-   }
-   else if ((GET_IRQ(sar_actflag) || GET_CPU(sar_actflag)) && WANT_PER_PROC(flags)) {
-      int nr_cpu;
-
-      nr_cpu = count_bits(cpu_bitmap, sizeof(cpu_bitmap));
-      if (nr_cpu > 1)
-	 dis_hdr = 1;
-   }
+   /* Determine if a stat line header has to be displayed */
+   check_line_hdr(&dis_hdr);
 
    lines = rows = get_win_height();
 
@@ -1504,7 +1434,7 @@ void read_stats(void)
     */
 
    /* Perform required allocations */
-   allocate_structures(USE_SADC);
+   allocate_structures();
 
    /* Print report header */
    print_report_hdr(flags, &rectime, &file_hdr);
@@ -1517,7 +1447,7 @@ void read_stats(void)
       write_stats_startup(0);
 
    /* Save the first stats collected. Will be used to compute the average */
-   copy_structures(2, 0, USE_SADC);
+   copy_structures(2, 0);
 
    /* Main loop */
    do {
@@ -1539,7 +1469,7 @@ void read_stats(void)
       if (file_stats[curr].record_type == R_LAST_STATS) {
 	 /* File rotation is happening: re-read header data sent by sadc */
 	 read_header_data();
-	 allocate_structures(USE_SADC);
+	 allocate_structures();
       }
 
       if (count > 0)
@@ -1563,7 +1493,6 @@ int main(int argc, char **argv)
 {
    int opt = 1, args_idx = 2;
    int fd[2];
-   unsigned long pid = 0;
    char from_file[MAX_FILE_LEN], to_file[MAX_FILE_LEN];
    char ltemp[20];
 
@@ -1647,52 +1576,6 @@ int main(int argc, char **argv)
 	 flags |= S_F_I_OPTION;
       }
 
-      else if (!strcmp(argv[opt], "-x") || !strcmp(argv[opt], "-X")) {
-	 if (argv[++opt]) {
-	    if (!strcmp(argv[opt], K_SADC)) {
-	       strcpy(ltemp, K_SELF);
-	       opt++;
-	    }
-	    else if (!strcmp(argv[opt], K_ALL)) {
-	       if (args_idx < MAX_ARGV_NR - 7) {
-		  salloc(args_idx++, argv[opt - 1]);	/* "-x" or "-X" */
-		  salloc(args_idx++, K_ALL);
-		  if (!strcmp(argv[opt - 1], "-x"))
-		     sar_actflag |= A_PID;
-		  else
-		     sar_actflag |= A_CPID;
-	       }
-	       opt++;
-	       continue;	/* Next option */
-	    }
-	    else {
-	       if (!strcmp(argv[opt], K_SELF)) {
-		  pid = getpid();
-		  opt++;
-	       }
-	       else if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
-		  usage(argv[0]);
-	       else {
-		  pid = atol(argv[opt++]);
-		  if (pid < 1)
-		     usage(argv[0]);
-	       }
-	       sprintf(ltemp, "%ld", pid);
-	    }
-
-	    if (args_idx < MAX_ARGV_NR - 7) {
-	       if (!strcmp(argv[opt - 2], "-x"))
-		  sar_actflag |= A_PID;
-	       else
-		  sar_actflag |= A_CPID;
-	       salloc(args_idx++, argv[opt - 2]);	/* "-x" or "-X" */
-	       salloc(args_idx++, ltemp);
-	    }
-	 }
-	 else
-	    usage(argv[0]);
-      }
-
       else if (!strcmp(argv[opt], "-n")) {
 	 if (argv[++opt]) {
 	    /* Parse option -n */
@@ -1739,6 +1622,9 @@ int main(int argc, char **argv)
        ((interval < 0) && !from_file[0] && !to_file[0]))
       set_default_file(&rectime, from_file);
 
+   if (tm_start.use && tm_end.use && (tm_end.tm_hour < tm_start.tm_hour))
+      tm_end.tm_hour += 24;
+
    /*
     * Check option dependencies
     */
@@ -1767,12 +1653,6 @@ int main(int argc, char **argv)
 	 count = 1;
    }
 
-   /* -x and -X options ignored when writing to a file */
-   if (to_file[0]) {
-      sar_actflag &= ~A_PID;
-      sar_actflag &= ~A_CPID;
-   }
-
    /* Default is CPU activity... */
    if (!sar_actflag)
       /*
@@ -1837,10 +1717,8 @@ int main(int argc, char **argv)
 
       /* Flags to be passed to sadc */
       salloc(args_idx++, "-z");
-      if (GET_ONE_IRQ(sar_actflag))
-	 salloc(args_idx++, "-I");
-      if (GET_DISK(sar_actflag))
-	 salloc(args_idx++, "-d");
+      salloc(args_idx++, "-I");
+      salloc(args_idx++, "-d");
 
       /* Outfile arg */
       if (to_file[0])
